import logging
from typing import TYPE_CHECKING

from PyQt6.QtCore import QEvent
from PyQt6.QtWidgets import QPushButton, QWidget, QHBoxLayout
from feeluown.gui.widgets.selfpaint_btn import TriagleButton

from feeluown.player import State
from feeluown.excs import ProviderIOError
from feeluown.utils.aio import run_fn
from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.widgets import PlayPauseButton, PlayNextButton, PlayPreviousButton
from feeluown.gui.helpers import resize_font
from feeluown.i18n import t

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


class LyricButton(TextButton):
    def __init__(self, app: "GuiApp", **kwargs):
        kwargs.setdefault("height", 16)
        super().__init__(t("lyric-button"), **kwargs)
        self._app = app

        font_size = 9
        self.setCheckable(True)
        self.clicked.connect(self._toggle_lyric_window)
        font = self.font()
        font.setPixelSize(font_size)
        self.setFont(font)

        self._app.ui.lyric_window.installEventFilter(self)

    def _toggle_lyric_window(self):
        lyric_window = self._app.ui.lyric_window
        if lyric_window.isVisible():
            lyric_window.hide()
        else:
            lyric_window.show()

    def eventFilter(self, _, event):
        """Event filter for lyric window"""
        if event.type() == QEvent.Type.Show:
            self.setChecked(True)
        elif event.type() == QEvent.Type.Hide:
            self.setChecked(False)
        return False


class WatchButton(TextButton):
    def __init__(self, app: "GuiApp", *args, **kwargs):
        super().__init__("♨", *args, **kwargs)
        self._app = app

        self.setToolTip(
            t("watch-mode-tooltip"),
        )

        self.setCheckable(True)
        self._setup_ui()
        self.clicked.connect(self.on_clicked)

    def on_clicked(self):
        self._app.playlist.watch_mode = not self._app.playlist.watch_mode
        # Assume that playlist.watch_mode will only be changed by WatchButton
        if self._app.playlist.watch_mode is True:
            self.setChecked(True)
        else:
            self.setChecked(False)

    def _setup_ui(self):
        font = self.font()
        resize_font(font, -2)
        self.setFont(font)

    def showEvent(self, e):
        self.setChecked(self._app.playlist.watch_mode)
        super().showEvent(e)


class FavButton(QPushButton):
    def __init__(self, app: "GuiApp", size=(13, 13), parent=None):
        super().__init__(parent=parent)
        self._app = app
        self.setCheckable(True)
        self.setFixedSize(*size)

        self.clicked.connect(self.toggle_liked)
        self.toggled.connect(self.on_toggled)
        self.setObjectName("like_btn")
        self.setToolTip(t("local-liked-tracks-add"))

        self._model = None

    def set_model(self, model):
        self._model = model
        self.setDisabled(model is None)
        if model is not None:
            self.setChecked(self.already_in_library(model))

    def toggle_liked(self):
        coll_library = self._app.coll_mgr.get_coll_library()
        model = self._model
        if self.already_in_library(model):
            coll_library.remove(model)
            self._app.show_msg(t("local-liked-tracks-removed"))
        else:
            coll_library.add(model)
            self._app.show_msg(t("local-liked-tracks-added"))

    def on_toggled(self):
        if self.already_in_library(self._model):
            self.setToolTip(t("local-liked-tracks-add"))
        else:
            self.setToolTip(t("local-liked-tracks-remove"))

    def already_in_library(self, model):
        coll_library = self._app.coll_mgr.get_coll_library()
        return model in coll_library.models


class LikeButton(FavButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFixedSize(15, 15)
        self._app.playlist.song_changed_v2.connect(self.on_song_changed)

    def on_song_changed(self, song, media):
        self.set_model(song)


class SongMVTextButton(TextButton):
    def __init__(self, app: "GuiApp", song=None, text="MV", **kwargs):
        super().__init__(text, **kwargs)
        self._app = app
        self._song = None
        self._mv = None

        self.bind_song(song)
        self.setDisabled(True)
        self.clicked.connect(self.on_clicked)

    def on_clicked(self):
        if self._mv is not None:
            self._app.watch_mgr.play_video(self._mv)

    def bind_song(self, song):
        if song != self._song:
            self._song = song
            self._mv = None

    async def get_mv(self):
        if self._song is None:
            return None

        try:
            mv = await run_fn(self._app.library.song_get_mv, self._song)
        except ProviderIOError:
            logger.exception("get song mv info failed")
            mv = None
            self.setEnabled(False)
        else:
            if mv is None:
                self.setEnabled(False)
            else:
                self.setToolTip(mv.title)
                self.setEnabled(True)
        self._mv = mv
        return self._mv


class NowplayingMVTextButton(SongMVTextButton):
    def __init__(self, app: "GuiApp", parent=None, **kwargs):
        super().__init__(app, song=None, parent=parent, **kwargs)

        self.setObjectName("mv_btn")
        self._app.playlist.song_mv_changed.connect(
            self.on_song_mv_changed, aioqueue=True
        )

    def on_song_mv_changed(self, _, mv):
        self.setEnabled(mv is not None)
        self._mv = mv


class MediaButtonsV2(QWidget):
    def __init__(self, app: "GuiApp", button_width=30, spacing=0, parent=None):
        super().__init__(parent=parent)
        self._app = app

        self.previous_btn = PlayPreviousButton(parent=self, length=button_width)
        self.pp_btn = PlayPauseButton(parent=self, length=button_width)
        self.next_btn = PlayNextButton(parent=self, length=button_width)
        self.toggle_video_btn = TriagleButton(length=button_width)
        self.toggle_video_btn.hide()
        self.pp_btn.setCheckable(True)
        self.toggle_video_btn.setToolTip(t("show-track-movie"))

        self._layout = QHBoxLayout(self)
        self._layout.setSpacing(spacing)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self.previous_btn)
        self._layout.addWidget(self.pp_btn)
        self._layout.addWidget(self.next_btn)
        self._layout.addWidget(self.toggle_video_btn)

        self.next_btn.clicked.connect(self._app.playlist.next)
        self.previous_btn.clicked.connect(self._app.playlist.previous)
        self.pp_btn.clicked.connect(self._app.player.toggle)
        self._app.player.state_changed.connect(
            self._on_player_state_changed, aioqueue=True
        )
        self._app.player.video_format_changed.connect(
            self.on_video_format_changed, aioqueue=True
        )

    def _on_player_state_changed(self, state):
        self.pp_btn.setChecked(state == State.playing)

    def on_video_format_changed(self, video_format):
        if video_format is None:
            self.toggle_video_btn.hide()
        else:
            self.toggle_video_btn.show()



from typing import TYPE_CHECKING

import asyncio
import logging

from feeluown.excs import ProviderIOError
from feeluown.player import PlaylistMode
from feeluown.i18n import t

if TYPE_CHECKING:
    from feeluown.app import App

logger = logging.getLogger(__name__)


class FM:
    """FM

    We can activate/deactivate playlist fm mode throught this FM manager.
    We also listen to playlist mode changed signal, when playlist exit
    fm mode, we will do some cleaning so that fm is completely deactivated.

    .. note::

        In fact, FM and playlist are a whole. Seperating them brings
        more clear API, which also cause some problems. For example,
        others can deactivate fm mode through playlist or FM, which
        maybe a bit confusing.
    """

    def __init__(self, app: "App"):
        """
        :type app: feeluown.app.App
        """
        self._app = app

        self._activated = False
        self._is_fetching_songs = False
        self._fetch_songs_task_name = "fm-fetch-songs"
        self._fetch_songs_func = None  # fn(number_to_fetch)
        self._minimum_per_fetch = 3

        self._app.playlist.mode_changed.connect(self._on_playlist_mode_changed)

    def activate(self, fetch_songs_func, reset=True):
        """change playlist mode to fm

        :param fetch_songs_func: `func(minimum, *args, **kwargs): -> list`
            please ensure that fetch_songs_func can receive keyword arguments,
            we may send some keyword args(such as timeout) in the future.
            If exception occured in fetch_songs_func, it should raise
            :class:`feeluown.excs.ProviderIOError`.
        :param reset: reset the playlist and play the next song

        .. versionadded:: 3.7.11
           The *reset* parameter.
        """
        if self.is_active:
            logger.warning("fm already actiavted")
            return
        self._fetch_songs_func = fetch_songs_func
        self._app.playlist.eof_reached.connect(self._on_playlist_eof_reached)
        self._app.playlist.mode = PlaylistMode.fm
        if reset is True:
            self._app.playlist.clear()
            self._app.playlist.next()
            self._app.player.resume()
        logger.info("fm mode actiavted")

    def deactivate(self):
        """deactivate fm mode"""
        self._app.playlist.mode = PlaylistMode.normal

    @property
    def is_active(self):
        """if fm mode is still active

        We can only activate fm mode by using `activate` method, but we can
        deactivate it through `deactivate` method or `playlist.mode.setter`.
        """
        return self._app.playlist.mode is PlaylistMode.fm

    def _on_playlist_eof_reached(self):
        if self._is_fetching_songs:
            return

        self._is_fetching_songs = True
        task_spec = self._app.task_mgr.get_or_create(self._fetch_songs_task_name)
        task = task_spec.bind_blocking_io(
            self._fetch_songs_func, self._minimum_per_fetch
        )
        task.add_done_callback(self._on_songs_fetched)

    def _on_playlist_mode_changed(self, mode):
        if mode is PlaylistMode.fm:
            return
        self._on_playlist_fm_mode_exited()

    def _on_playlist_fm_mode_exited(self):
        """when playlist fm mode exited,"""
        self._app.playlist.eof_reached.disconnect(self._on_playlist_eof_reached)
        self._fetch_songs_func = None
        logger.info("fm mode deactivated")

    def _feed_playlist(self, songs):
        for song in songs:
            self._app.playlist.fm_add(song)
        self._app.playlist.next()

    def _on_songs_fetched(self, future):
        try:
            songs = future.result()
        except asyncio.CancelledError:
            logger.exception("fm-fetch-songs task is cancelled")
        except ProviderIOError:
            logger.exception("fm-fetch-songs io error")
        else:
            if len(songs) < self._minimum_per_fetch:
                logger.info(t("track-radio-not-enough", locale="en-US"))
                self._app.show_msg(t("track-radio-not-enough"))
                self.deactivate()
            else:
                self._feed_playlist(songs)
        finally:
            self._is_fetching_songs = False



from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt

from feeluown.gui.page_containers.scroll_area import ScrollArea
from feeluown.gui.widgets.labels import MessageLabel

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


async def render_error_message(app: "GuiApp", msg: str):
    label = MessageLabel(msg, MessageLabel.ERROR)
    label.setAlignment(Qt.AlignmentFlag.AlignCenter)
    app.ui.page_view.set_body(label)


async def render_scroll_area_view(req, view_cls):
    app: "GuiApp" = req.ctx["app"]
    view = view_cls(app)
    # Keep page rendering behavior consistent across pages that need a
    # scrollable body with custom widgets.
    scroll_area = ScrollArea()
    scroll_area.setWidget(view)
    app.ui.right_panel.set_body(scroll_area)
    await view.render()



from feeluown.app.gui_app import GuiApp
from feeluown.library import ModelType
from feeluown.utils.aio import run_fn
from feeluown.gui.page_containers.table import Renderer
from feeluown.gui.base_renderer import TabBarRendererMixin
from feeluown.library import (
    SupportsCurrentUserFavSongsReader,
    SupportsCurrentUserFavAlbumsReader,
    SupportsCurrentUserFavArtistsReader,
    SupportsCurrentUserFavPlaylistsReader,
    SupportsCurrentUserFavVideosReader,
)
from feeluown.utils.reader import create_reader
from feeluown.i18n import t
from .template import render_error_message


async def render(req, **kwargs):
    app: GuiApp = req.ctx["app"]
    ui = app.ui
    tab_index = int(req.query.get("tab_index", 0))
    pvd_ui = app.current_pvd_ui_mgr.get()
    if pvd_ui is None:
        return await render_error_message(app, t("provider-unknown-cannot-view"))

    ui.right_panel.set_body(ui.right_panel.scrollarea)
    table_container = ui.right_panel.table_container
    renderer = MyFavRenderer(app, tab_index, pvd_ui.provider)
    await table_container.set_renderer(renderer)


class MyFavRenderer(Renderer, TabBarRendererMixin):
    def __init__(self, app, tab_index, provider):
        self._app = app
        self._provider = provider
        self.tab_index = tab_index
        self.tabs = self.default_tabs()

    async def render(self):
        self.meta_widget.show()
        self.meta_widget.title = t("my-favorite-title")
        self.render_tab_bar()
        await self.render_models()

    async def render_models(self):
        # pylint: disable=too-many-branches
        _, mtype, show_handler = self.tabs[self.tab_index]
        media_type = ""
        reader = create_reader([])
        if mtype is ModelType.song:
            if isinstance(self._provider, SupportsCurrentUserFavSongsReader):
                reader = await run_fn(self._provider.current_user_fav_create_songs_rd)
            else:
                media_type = "track"
        elif mtype is ModelType.album:
            if isinstance(self._provider, SupportsCurrentUserFavAlbumsReader):
                reader = await run_fn(self._provider.current_user_fav_create_albums_rd)
            else:
                media_type = "album"
        elif mtype is ModelType.artist:
            if isinstance(self._provider, SupportsCurrentUserFavArtistsReader):
                reader = await run_fn(self._provider.current_user_fav_create_artists_rd)
            else:
                media_type = "singer"
        elif mtype is ModelType.playlist:
            if isinstance(self._provider, SupportsCurrentUserFavPlaylistsReader):
                reader = await run_fn(
                    self._provider.current_user_fav_create_playlists_rd
                )  # noqa
            else:
                media_type = "playlist"
        else:
            if isinstance(self._provider, SupportsCurrentUserFavVideosReader):
                reader = await run_fn(self._provider.current_user_fav_create_videos_rd)
            else:
                media_type = "video"
        if media_type:
            return await render_error_message(
                self._app,
                t(
                    "provider-missing-favorite",
                    providerName=self._provider.name,
                    mediaType=media_type,
                ),
            )
        else:
            show_handler(reader)

    def render_by_tab_index(self, tab_index):
        self._app.browser.goto(page="/my_fav", query={"tab_index": tab_index})



import asyncio
import logging
import random
from enum import IntEnum, Enum
from typing import Optional, TYPE_CHECKING
from threading import Lock

from feeluown.excs import ProviderIOError
from feeluown.utils import aio
from feeluown.utils.aio import run_fn, run_afn
from feeluown.utils.dispatch import Signal
from feeluown.utils.utils import DedupList
from feeluown.library import (
    MediaNotFound,
    SongModel,
    ModelType,
    VideoModel,
    ModelNotFound,
    BriefSongModel,
)
from feeluown.media import Media
from feeluown.i18n import t
from .metadata_assembler import MetadataAssembler

if TYPE_CHECKING:
    from feeluown.app import App

logger = logging.getLogger(__name__)

TASK_SET_CURRENT_MODEL = "playlist.set_current_model"
TASK_PLAY_MODEL = "playlist.play_model"
TASK_PREPARE_MEDIA = "playlist.prepare_media"


class PlaybackMode(IntEnum):
    """
    Playlist playback mode.

    .. deprecated:: 3.8.12
        Please use PlaylistRepeatMode and PlaylistShuffleMode instead.
    """

    one_loop = 0  #: One Loop
    sequential = 1  #: Sequential
    loop = 2  #: Loop
    random = 3  #: Random


class PlaylistRepeatMode(Enum):
    none = "none"
    one = "one"
    all = "all"


class PlaylistShuffleMode(Enum):
    """
    Windows and macOS has multiple shuffle modes.
    """

    off = "off"
    songs = "songs"


class PlaylistMode(IntEnum):
    """playlist mode

    **What is FM mode?**

    In FM mode, playlist's playback_mode is unchangeable, it will
    always be sequential. When playlist has no more song,
    the playlist hopes someone(we call it ``FM`` here) will:
    1. catch the ``eof_reached`` signal
    2. add news songs to playlist by using ``fm_add`` method

    **When will playlist exit FM mode?**

    If user manually play a song, playlist will exit FM mode, at the
    same time, playlist will:
    1. clear itself
    2. change to normal mode
    3. set current song to the song
    """

    normal = 0  #: Normal
    fm = 1  #: FM mode


class PlaylistPlayModelStage(IntEnum):
    prepare_media = 10
    find_standby_by_mv = 20
    find_standby = 30
    prepare_metadata = 40
    load_media = 50


class Playlist:
    def __init__(
        self,
        app: "App",
        songs=None,
        playback_mode=PlaybackMode.loop,
        audio_select_policy="hq<>",
    ):
        """
        :param songs: list of :class:`feeluown.library.SongModel`
        :param playback_mode: :class:`feeluown.player.PlaybackMode`
        """
        self._app = app
        self._metadata_mgr = MetadataAssembler(app)

        #: init playlist mode normal
        self._mode = PlaylistMode.normal

        #: playlist eof signal
        # playlist have no enough songs
        self.eof_reached = Signal()

        #: playlist mode changed signal
        self.mode_changed = Signal()
        #: playback mode before changed to fm mode
        self._normal_mode_playback_mode = playback_mode

        #: store value for ``current_song`` property
        self._current_song = None
        self._current_song_mv = None

        #: songs whose url is invalid
        self._bad_songs = DedupList()

        # A data structure to store the song list.
        self._songs = DedupList(songs or [])
        self._shuffled_songs = None
        self._queue = self._songs  # A pointer the the current song list.

        # Acquire this lock before changing _current_song or _songs.
        # NOTE: some methods emit some signal while holding the lock,
        #   the signal handler should pay much attention to avoid deadlock.
        #   One thing is that the signal handler should not call any method
        #   that requires the lock!!!
        self._queue_lock = Lock()

        self.audio_select_policy = audio_select_policy

        #: store value for ``playback_mode`` property
        self._playback_mode = playback_mode

        #: playback mode changed signal
        self.playback_mode_changed = Signal()
        self.song_changed = Signal()
        """current song changed signal

        The player will play the song after it receive the signal,
        when song is None, the player will stop playback.
        """
        self.song_changed_v2 = Signal()
        """current song chagned signal, v2

        emit(song, media)
        """
        self.song_mv_changed = Signal()  # emit(song, mv)

        #: When watch mode is on, playlist try to play the mv/video of the song
        self.watch_mode = False

        # .. versionadded:: 3.7.11
        #    The *songs_removed* and *songs_added* signal.
        self.songs_removed = Signal()  # (index, count)
        self.songs_added = Signal()  # (index, count)
        self.songs_reordered = Signal()  # ()
        # .. versionadded:: 3.9.0
        #    The *play_model_handling* signal.
        self.play_model_handling = Signal()
        # .. versionadded:: 4.1.7
        #    The *play_model_handling* signal.
        self.play_model_stage_changed = Signal()

        self._app.player.media_finished.connect(self._on_media_finished)
        self.song_changed.connect(self._on_song_changed)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        """set playlist mode"""
        if self._mode is not mode:
            # we should change _mode at the very end
            self._mode = mode
            if mode is PlaylistMode.fm:
                self._normal_mode_playback_mode = self.playback_mode
                self.playback_mode = PlaybackMode.sequential
            else:
                self.playback_mode = self._normal_mode_playback_mode
            self.mode_changed.emit(mode)
            logger.info("playlist mode changed to %s", mode)

    @property
    def repeat_mode(self):
        if self.playback_mode in (PlaybackMode.loop, PlaybackMode.random):
            return PlaylistRepeatMode.all
        elif self.playback_mode is PlaybackMode.one_loop:
            return PlaylistRepeatMode.one
        return PlaylistRepeatMode.none

    @repeat_mode.setter
    def repeat_mode(self, mode):
        if mode is PlaylistRepeatMode.all:
            if self.playback_mode not in (PlaybackMode.loop, PlaybackMode.random):
                self.playback_mode = PlaybackMode.loop
        elif mode is PlaylistRepeatMode.one:
            self.playback_mode = PlaybackMode.one_loop
        else:
            self.playback_mode = PlaybackMode.sequential

    @property
    def shuffle_mode(self):
        if self.playback_mode is PlaybackMode.random:
            return PlaylistShuffleMode.songs
        return PlaylistShuffleMode.off

    @shuffle_mode.setter
    def shuffle_mode(self, mode):
        if mode is PlaylistShuffleMode.songs:
            self.playback_mode = PlaybackMode.random
        else:
            self.playback_mode = PlaybackMode.loop

    def __len__(self):
        return len(self._queue)

    def __getitem__(self, index):
        """overload [] operator"""
        return self._queue[index]

    def mark_as_bad(self, song):
        if song in self._queue and song not in self._bad_songs:
            self._bad_songs.append(song)

    def is_bad(self, song):
        return song in self._bad_songs

    def _add(self, song):
        """
        Requires: acquire `_queue_lock` before calling this method.
        """
        if song in self._queue:
            return
        self._queue_append(song)
        length = len(self._queue)
        self.songs_added.emit(length - 1, 1)

    @property
    def _queue_is_shuffled_songs(self):
        return self._queue is self._shuffled_songs

    def _queue_insert(self, index, song):
        self._queue.insert(index, song)
        if self._queue_is_shuffled_songs:
            self._songs.append(song)

    def _queue_append(self, song):
        self._queue.append(song)
        if self._queue_is_shuffled_songs:
            self._songs.append(song)

    def _queue_remove(self, song):
        self._queue.remove(song)
        if self._queue_is_shuffled_songs:
            self._songs.remove(song)

    def _queue_clear(self):
        self._queue.clear()
        if self._queue_is_shuffled_songs:
            self._songs.clear()

    def set_models(self, models, next_=False, fm=False):
        """
        .. versionadded: v3.7.13
        """
        self.clear()
        self.batch_add(models)
        if fm is False:
            self.mode = PlaylistMode.normal
        else:
            self.mode = PlaylistMode.fm
        if next_ is True:
            self.next()

    def batch_add(self, models):
        """
        .. versionadded: v3.7.13
        """
        unique_models = DedupList(models)
        nonexisting_models = []
        for model in unique_models:
            if model not in self._queue:
                nonexisting_models.append(model)

        with self._queue_lock:
            start_index = len(self._queue)
            self._songs.extend(nonexisting_models)
            if self._queue_is_shuffled_songs:
                random.shuffle(nonexisting_models)
                self._shuffled_songs.extend(nonexisting_models)
            end_index = len(self._queue)
            self.songs_added.emit(start_index, end_index - start_index)

    def add(self, song):
        """add song to playlist

        Theoretically, when playlist is in FM mode, we should not
        change songs list manually(without ``fm_add`` method). However,
        when it happens, we exit FM mode.
        """
        if self._mode is PlaylistMode.fm:
            self.mode = PlaylistMode.normal
        with self._queue_lock:
            self._add(song)

    def fm_add(self, song):
        with self._queue_lock:
            self._fm_add_no_lock(song)

    def _fm_add_no_lock(self, song):
        """
        Only for unittest and internal usage.
        """
        self._add(song)

    def insert_after_current_song(self, song):
        """Insert song after current song

        Requires: acquire `_queue_lock` before calling this method.

        When current song is none, the song is appended.
        """
        if song in self._queue:
            return
        if self._mode is PlaylistMode.fm:
            self.mode = PlaylistMode.normal
        if self._current_song is None:
            self._add(song)
        else:
            index = self._queue.index(self._current_song)
            self._queue_insert(index + 1, song)
            self.songs_added.emit(index + 1, 1)

    def remove_no_lock(self, song):
        try:
            index = self._queue.index(song)
        except ValueError:
            logger.debug("Remove failed: {} not in playlist".format(song))
        else:
            if self._current_song is None:
                self._queue_remove(song)
            elif song == self._current_song:
                next_song = self._get_next_song_no_lock()
                # In random mode or when there is only one song left in the playlist,
                # the next song may be the same as the current one.
                if next_song == self.current_song:
                    # Should set current song immediately.
                    # Should not use set_current_song, because it is an async task.
                    self.set_current_song_none()
                    self._queue_remove(song)
                    new_next_song = self._get_next_song_no_lock()
                    self.set_existing_song_as_current_song(new_next_song)
                elif next_song is None and self.mode is PlaylistMode.fm:
                    # The caller should not remove the current song when it
                    # is the last song in fm mode.
                    logger.error(
                        "Can't remove the last song in fm mode, will play next"
                    )
                    self._next_no_lock()
                    return
                else:
                    self._queue_remove(song)
                    self.set_existing_song_as_current_song(next_song)
            else:
                self._queue_remove(song)
            self.songs_removed.emit(index, 1)
            logger.debug("Remove {} from player playlist".format(song))
        if song in self._bad_songs:
            self._bad_songs.remove(song)

    def remove(self, song):
        """Remove song from playlist. O(n)

        If song is current song, remove the song and play next. Otherwise,
        just remove it.
        """
        with self._queue_lock:
            self.remove_no_lock(song)

    def _replace_song_no_lock(self, model, umodel):
        index = self._queue.index(model)
        self._queue_insert(index + 1, umodel)
        self.songs_added.emit(index + 1, 1)
        if self.current_song == model:
            self.set_current_song_none()
        self._queue_remove(model)
        self.songs_removed.emit(index, 1)

    def clear(self):
        """remove all songs from playlists"""
        with self._queue_lock:
            if self.current_song is not None:
                self.set_current_song_none()
            length = len(self._queue)
            self._queue_clear()
            if length > 0:
                self.songs_removed.emit(0, length)
            self._bad_songs.clear()

    def list(self):
        """Get all songs in playlists"""
        return self._queue

    def list_unshuffled(self):
        """Get all songs in original order"""
        if self._shuffled_songs is not None:
            return list(self._songs)
        return list(self._queue)

    @property
    def playback_mode(self):
        return self._playback_mode

    @playback_mode.setter
    def playback_mode(self, playback_mode):
        if self._mode is PlaylistMode.fm:
            if playback_mode is not PlaybackMode.sequential:
                logger.warning("can't set playback mode to others in fm mode")
                return

        current_mode = self._playback_mode
        if current_mode is playback_mode:
            return

        if (
            current_mode is not PlaybackMode.random
            and playback_mode is PlaybackMode.random
        ):
            self._enter_shuffle_mode()
        elif (
            current_mode is PlaybackMode.random
            and playback_mode is not PlaybackMode.random
        ):
            self._leave_shuffle_mode()

        self._playback_mode = playback_mode
        self.playback_mode_changed.emit(self.playback_mode)

    def _enter_shuffle_mode(self):
        with self._queue_lock:
            assert self._shuffled_songs is None
            self._shuffled_songs = self._queue.copy()
            random.shuffle(self._shuffled_songs)
            self._queue = self._shuffled_songs
            self.songs_reordered.emit(0, len(self._queue))

    def _leave_shuffle_mode(self):
        with self._queue_lock:
            assert self._shuffled_songs is not None
            self._queue = self._songs
            self.songs_reordered.emit(0, len(self._queue))
            self._shuffled_songs = None

    def _get_good_song(self, base=0, random_=False, direction=1, loop=True):
        """Get a good song from playlist.

        Requires: acquire `_queue_lock` before calling this method.

        :param base: base index
        :param random_: random strategy or not
        :param direction: forward if > 0 else backward
        :param loop: regard the song list as a loop

        >>> from unittest import mock
        >>> pl = Playlist(mock.Mock(), [1, 2, 3])
        >>> pl._get_good_song()
        1
        >>> pl._get_good_song(base=1)
        2
        >>> pl._bad_songs = [2]
        >>> pl._get_good_song(base=1, direction=-1)
        1
        >>> pl._get_good_song(base=1)
        3
        >>> pl._bad_songs = [1, 2, 3]
        >>> pl._get_good_song()
        """
        if not self._queue or len(self._queue) <= len(self._bad_songs):
            logger.debug("No good song in playlist.")
            return None

        good_songs = []
        if direction > 0:
            if loop is True:
                song_list = self._queue[base:] + self._queue[0:base]
            else:
                song_list = self._queue[base:]
        else:
            if loop is True:
                song_list = self._queue[base::-1] + self._queue[:base:-1]
            else:
                song_list = self._queue[base::-1]
        for song in song_list:
            if song not in self._bad_songs:
                good_songs.append(song)
        if not good_songs:
            return None
        if random_:
            return random.choice(good_songs)
        else:
            return good_songs[0]

    def _get_next_song_no_lock(self):
        """
        Requires: acquire `_queue_lock` before calling this method.
        """
        assert self._queue_lock.locked()

        if self.current_song is None:
            return self._get_good_song()

        current_index = self._queue.index(self.current_song)
        is_last_song = current_index == len(self._queue) - 1
        if is_last_song and self.playback_mode == PlaybackMode.sequential:
            next_song = None
        else:
            if is_last_song:
                base_index = 0
            else:
                base_index = current_index + 1
            loop = self.playback_mode != PlaybackMode.sequential
            next_song = self._get_good_song(base=base_index, loop=loop)
        return next_song

    @property
    def next_song(self):
        """next song for player, calculated based on playback_mode"""
        # If there is no song currently playing,
        # find the first playable song in the list.
        with self._queue_lock:
            return self._get_next_song_no_lock()

    def _get_previous_song_no_lock(self):
        if self.current_song is None:
            return self._get_good_song(base=-1, direction=-1)

        current_index = self._queue.index(self.current_song)
        previous_song = self._get_good_song(base=current_index - 1, direction=-1)
        return previous_song

    @property
    def previous_song(self):
        """previous song for player to play

        NOTE: not the last played song
        """
        with self._queue_lock:
            return self._get_previous_song_no_lock()

    async def a_next(self):
        self.next()

    def _next_no_lock(self):
        """
        Requires: acquire `_queue_lock` before calling this method.
        Only for unittest and internal usage.
        """
        next_song = self._get_next_song_no_lock()
        if next_song is None:
            logger.debug("No next song in playlist, emit eof_reached.")
            self.eof_reached.emit()
            return None
        return self.set_existing_song_as_current_song(next_song)

    def next(self) -> Optional[asyncio.Task]:
        """
        Why _queue_lock is needed? Think about the following scenario:

            [A, B, C, D] is the playlist, and the current song is B.

            Timeline    t1             t2              t3      t4
            User:      play_next                     play D
            Player:                next_song is C

            The expected song to play is D. So lock is needed here.
        """
        with self._queue_lock:
            return self._next_no_lock()

    def _on_media_finished(self):
        # Play next model when current media is finished.
        if self.playback_mode == PlaybackMode.one_loop:
            with self._queue_lock:
                return self.set_existing_song_as_current_song(self.current_song)
        else:
            self.next()

    def _on_song_changed(self, song):
        self._app.task_mgr.run_afn_preemptive(self._fetch_current_song_mv, song)

    async def _fetch_current_song_mv(self, song):
        if song is None:
            self._current_song_mv = None
        else:
            try:
                mv = await run_fn(self._app.library.song_get_mv, song)
            except ProviderIOError:
                logger.exception("get song mv info failed")
                self._current_song_mv = None
            else:
                self._current_song_mv = mv
        self.song_mv_changed.emit(song, self._current_song_mv)

    def previous(self) -> Optional[asyncio.Task]:
        """return to the previous song in playlist"""
        with self._queue_lock:
            song = self._get_previous_song_no_lock()
            return self.set_existing_song_as_current_song(song)

    @property
    def current_song(self) -> Optional[SongModel]:
        """Current song

        return None if there is no current song
        """
        return self._current_song

    @current_song.setter
    def current_song(self, song: Optional[SongModel]):
        self.set_current_song(song)

    @property
    def current_song_mv(self) -> Optional[VideoModel]:
        return self._current_song_mv

    async def a_set_current_song(self, song):
        """Set the `song` as the current song.

        If the song is bad, then this will try to use a standby in Playlist.normal mode.
        """
        if song is None:
            self.set_current_song_none()
            return None

        if self.mode is PlaylistMode.fm and song not in self._queue:
            self.mode = PlaylistMode.normal

        target_song = song  # The song to be set.
        media = None  # The corresponding media to be set.
        try:
            self.play_model_stage_changed.emit(PlaylistPlayModelStage.prepare_media)
            media = await self._app.task_mgr.run_afn_preemptive(
                self._prepare_media,
                song,
                name=TASK_PREPARE_MEDIA,
            )
        except MediaNotFound as e:
            if e.reason is MediaNotFound.Reason.check_children:
                await self.a_set_current_song_children(song)
                return
        except ProviderIOError as e:
            # FIXME: This may cause infinite loop when the prepare media always fails
            logger.error(f"prepare media failed: {e}, try next song")
            run_afn(self.a_next)
            return
        except Exception as e:  # noqa
            # When the exception is unknown, we mark the song as bad.
            self._app.show_msg(t("track-url-fetch-failed", errorMessage=e))
            logger.exception("prepare media failed due to unknown error")
        else:
            assert media, "media must not be empty"

        # The song has no media, try to find and use standby unless it is in fm mode.
        if media is None:
            if self._app.config.ENABLE_MV_AS_STANDBY:
                self.play_model_stage_changed.emit(
                    PlaylistPlayModelStage.find_standby_by_mv
                )
                media = await self._prepare_mv_media(song)

            if media:
                self._app.show_msg(t("track-fallback-music-video"))
            else:
                self._app.show_msg(t("track-fallback-no-music-video"))
                logger.info(f"no media found for {song}, mark it as bad")
                self.mark_as_bad(song)
                self.play_model_stage_changed.emit(PlaylistPlayModelStage.find_standby)
                target_song, media = await self.find_and_use_standby(song)

        metadata = None
        if media is not None:
            self.play_model_stage_changed.emit(PlaylistPlayModelStage.prepare_metadata)
            metadata = await self._metadata_mgr.prepare_for_song(target_song)
        self.play_model_stage_changed.emit(PlaylistPlayModelStage.load_media)
        self.set_current_song_with_media(target_song, media, metadata)

    async def a_set_current_song_children(self, song):
        # TODO: maybe we can just add children to playlist?
        self._app.show_msg(f"{song} 的播放资源在孩子节点上，将孩子节点添加到播放列表")
        self.mark_as_bad(song)
        logger.info(f"{song} has children, replace the current playlist")
        song = await run_fn(self._app.library.song_upgrade, song)
        if song.children:
            self.batch_add(song.children)
            await self.a_set_current_song(song.children[0])
        else:
            run_afn(self.a_next)
        return

    async def find_and_use_standby(self, song):
        self._app.show_msg(t("track-standby-try", track=song))
        logger.info(f"try to find standby from other providers for {song}")
        standby_candidates = await self._app.library.a_list_song_standby_v2(
            song, self.audio_select_policy
        )
        if standby_candidates:
            standby, media = standby_candidates[0]
            logger.info(f"song standby was found in {standby.source} ✅")
            self._app.show_msg(
                t("track-standby-found", track=song, standby=standby.source)
            )
            # Insert the standby song after the song
            # TODO: Perhaps we can optimize here using the `self.insert` function?
            with self._queue_lock:
                if song in self._queue and standby not in self._queue:
                    index = self._queue.index(song)
                    self._queue_insert(index + 1, standby)
                    self.songs_added.emit(index + 1, 1)
            return standby, media

        logger.info(f"{song} song standby not found")
        self._app.show_msg(t("track-standby-unavailable", track=song))
        return song, None

    def set_current_song_with_media(self, song, media, metadata=None):
        if song is None:
            self.set_current_song_none()
            return
        # Add it to playlist if song not in playlist.
        with self._queue_lock:
            self.insert_after_current_song(song)
            self._current_song = song
            # TODO: There might be a problem here.
            # For example, how do we keep `current_song` consistent with `media`?
            self.song_changed.emit(song)
            self.song_changed_v2.emit(song, media)
        if media is None:
            self._app.show_msg(t("track-skip-to-next"))
            run_afn(self.a_next)
        else:
            kwargs = {}
            if not self._app.has_gui:
                kwargs["video"] = False
            # TODO: set artwork field
            self._app.player.play(media, metadata=metadata, **kwargs)

    def set_current_song_none(self):
        """A special case of `set_current_song_with_media`."""
        self._current_song = None
        self.song_changed.emit(None)
        self.song_changed_v2.emit(None, None)
        self._app.player.stop()

    async def _prepare_media(self, song):
        if self.watch_mode is True:
            mv_media = await self._prepare_mv_media(song)
            if mv_media:
                return mv_media
            self._app.show_msg(t("music-video-not-avaliable"))
        return await aio.run_fn(
            self._app.library.song_prepare_media,
            song,
            self.audio_select_policy,
        )

    async def _prepare_mv_media(self, song) -> Optional[Media]:
        try:
            mv_media = await run_fn(
                self._app.library.song_prepare_mv_media,
                song,
                self._app.config.VIDEO_SELECT_POLICY,
            )
        except MediaNotFound:
            mv_media = None
        except Exception as e:  # noqa
            mv_media = None
            logger.exception(f"fail to get {song} mv: {e}")
        return mv_media

    async def a_set_current_model(self, model):
        """
        TODO: handle when model is a song

        .. versionadded: 3.7.13
        """
        if model is None:
            self._app.player.stop()
            return
        if isinstance(model, BriefSongModel):
            return await self.a_set_current_song(model)

        video = model
        try:
            media = await aio.run_fn(
                self._app.library.video_prepare_media,
                video,
                self._app.config.VIDEO_SELECT_POLICY,
            )
        except MediaNotFound:
            self._app.show_msg(t("playback-url-unavailable"))
        else:
            metadata = await self._metadata_mgr.prepare_for_video(video)
            kwargs = {}
            if not self._app.has_gui:
                kwargs["video"] = False
            self._app.player.play(media, metadata=metadata, **kwargs)

    async def a_play_model(self, model):
        """
        .. versionadded: 4.1.7
        """
        # Stop the player so that user know the action is working.
        self._app.player.stop()
        if model is None:
            return
        self.play_model_handling.emit()
        if ModelType(model.meta.model_type) is ModelType.song:
            fn = self.a_set_current_song
            upgrade_fn = self._app.library.song_upgrade
        else:
            fn = self.a_set_current_model
            upgrade_fn = self._app.library.video_upgrade

        try:
            # Try to upgrade the model.
            umodel = await aio.run_fn(upgrade_fn, model)
        except ModelNotFound:
            pass
        except Exception as e:  # noqa
            logger.exception(f"upgrade model({model}) failed")
            self._app.alert_mgr.on_exception(e)
        else:
            # Replace the brief model with the upgraded model
            # when user try to play a brief model that is already in the playlist.
            if isinstance(model, BriefSongModel) and not isinstance(model, SongModel):
                with self._queue_lock:
                    if model in self._queue:
                        self._replace_song_no_lock(model, umodel)
            model = umodel

        try:
            await self._app.task_mgr.run_afn_preemptive(
                fn, model, name=TASK_SET_CURRENT_MODEL
            )
        except:  # noqa
            logger.exception(f"play model({model}) failed")
        else:
            self._app.player.resume()
            logger.info(f"play a model ({model}) succeed")

    """
    Sync methods.

    Currently, playlist has both async and sync methods to keep backward
    compatibility. Sync methods will be replaced by async methods in the end.
    Sync methods just wrap the async method.
    """

    def play_model(self, model):
        """Set current model and play it

        .. versionadded: 3.7.14
        """
        self._app.task_mgr.run_afn_preemptive(
            self.a_play_model, model, name=TASK_PLAY_MODEL
        )

    def set_current_model(self, model) -> asyncio.Task:
        """
        .. versionadded: 3.7.13
        """
        return self._app.task_mgr.run_afn_preemptive(
            self.a_set_current_model,
            model,
            name=TASK_SET_CURRENT_MODEL,
        )

    def set_existing_song_as_current_song(self, song):
        """
        Requires: acquire `_queue_lock` before calling this method.
        """
        self._current_song = song
        return self.set_current_song(song)

    def set_current_song(self, song):
        """
        .. versionadded:: 3.7.11
           The method is added to replace current_song.setter.
        """
        return self._app.task_mgr.run_afn_preemptive(
            self.a_set_current_song, song, name=TASK_SET_CURRENT_MODEL
        )



import logging
import os
import sys
import warnings

from feeluown import logger_config
from feeluown.app import App
from feeluown.consts import (
    HOME_DIR, USER_PLUGINS_DIR, DATA_DIR,
    CACHE_DIR, USER_THEMES_DIR, SONG_DIR, COLLECTIONS_DIR
)

logger = logging.getLogger(__name__)


def ensure_dirs():
    for d in (HOME_DIR,
              DATA_DIR,
              USER_THEMES_DIR,
              USER_PLUGINS_DIR,
              CACHE_DIR,
              SONG_DIR,
              COLLECTIONS_DIR):
        if not os.path.exists(d):
            os.mkdir(d)


def setup_config(args, config):
    config.DEBUG = args.debug or config.DEBUG
    config.VERBOSE = args.verbose or config.VERBOSE
    config.MPV_AUDIO_DEVICE = args.mpv_audio_device or config.MPV_AUDIO_DEVICE
    config.LOG_TO_FILE = bool(args.log_to_file or config.LOG_TO_FILE or
                              os.getenv('FUO_LOG_TO_FILE'))

    if args.cmd is not None:
        config.MODE = App.CliMode
        # Always log to file in cli mode because logs may pollute the output.
        config.LOG_TO_FILE = True
    else:
        if not args.no_window:
            try:
                import PyQt6  # noqa, pylint: disable=unused-import
            except ImportError:
                logger.warning('PyQt6 is not installed, fallback to daemon mode.')
            else:
                try:
                    from feeluown.utils.compat import QEventLoop  # noqa
                except ImportError:
                    logger.warning('no QEventLoop, fallback to daemon mode.')
                else:
                    config.MODE |= App.GuiMode
        if not args.no_server:
            config.MODE |= App.DaemonMode


def setup_logger(config):
    if config.DEBUG:
        verbose = 3
    else:
        verbose = config.VERBOSE
    logger_config(verbose=verbose, to_file=config.LOG_TO_FILE)
    # Show deprecation warning when user does not set it.
    if not sys.warnoptions and verbose >= 2:
        warnings.simplefilter('default', DeprecationWarning)



from datetime import datetime

from PyQt6.QtWidgets import QAbstractItemView, QFrame, QVBoxLayout, QScrollArea

from feeluown.library import SearchType
from feeluown.utils.aio import run_afn
from feeluown.gui.components.overlay import AppOverlayContainer
from feeluown.gui.page_containers.table import TableContainer, Renderer
from feeluown.gui.widgets.img_card_list import ImgCardListDelegate
from feeluown.gui.widgets.songs import SongsTableView, ColumnsMode
from feeluown.gui.base_renderer import TabBarRendererMixin
from feeluown.gui.helpers import (
    BgTransparentMixin,
    unify_scroll_area_style,
    set_widget_bg_transparent,
)
from feeluown.gui.widgets.magicbox import KeySourceIn, KeyType
from feeluown.gui.widgets.header import LargeHeader, MidHeader
from feeluown.gui.widgets.accordion import Accordion
from feeluown.gui.widgets.labels import MessageLabel
from feeluown.utils.reader import create_reader
from feeluown.i18n import t


Tabs = [
    (t("track"), SearchType.so),
    (t("album"), SearchType.al),
    (t("musician"), SearchType.ar),
    (t("playlist"), SearchType.pl),
    (t("video"), SearchType.vi),
]


def get_tab_idx(search_type):
    for i, tab in enumerate(Tabs):
        if tab[1] == search_type:
            return i
    raise ValueError("unknown search type")


def create_search_result_view(app, song):
    q = f"{song.title} {song.artists_name}"
    body = SearchResultView(app, transparent_bg=False)
    view = AppOverlayContainer(app, body, parent=app, adhoc=True)

    source_in = app.browser.local_storage.get(KeySourceIn, None)
    run_afn(body.search_and_render, q, SearchType.so, source_in)
    return view


class SearchResultView(QScrollArea):
    """
    Usage:
        view = SearchResultView(app)
        await view.search_and_render(q, search_type, source_in)
    """

    def __init__(self, app, transparent_bg=True, parent=None):
        super().__init__(parent=parent)

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.Shape.NoFrame)
        unify_scroll_area_style(self)
        if transparent_bg:
            set_widget_bg_transparent(self)

        self.body = Body(app)
        self.setWidget(self.body)

    def fillable_bg_height(self):
        """Implement VFillableBg protocol"""
        return self.body.height() - self.body.accordion.height()

    async def search_and_render(self, q, search_type, source_in):
        await self.body.search_and_render(q, search_type, source_in)


class Body(QFrame, BgTransparentMixin):
    def __init__(self, app, **kwargs):
        super().__init__(**kwargs)

        self._app = app

        self.title = LargeHeader()
        self.hint = MessageLabel()
        self.accordion = Accordion()

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(20, 0, 20, 0)
        self._layout.setSpacing(0)
        self._layout.addSpacing(30)
        self._layout.addWidget(self.title)
        self._layout.addSpacing(10)
        self._layout.addWidget(self.hint)
        self._layout.addSpacing(10)
        self._layout.addWidget(self.accordion)
        self._layout.addStretch(0)

    async def search_and_render(self, q, search_type, source_in):
        # pylint: disable=too-many-locals,too-many-statements
        view = self
        app = self._app

        self.title.setText(t("track-search", keyword=q))

        tab_index = get_tab_idx(search_type)
        succeed = 0
        start = datetime.now()
        is_first = True  # Is first search result.
        if source_in is not None:
            source_count = len(source_in)
        else:
            source_count = len(app.library.list())
        hint_msgs = [t("track-searching", providerCount=source_count)]
        view.hint.show_msg("\n".join(hint_msgs))
        async for result in app.library.a_search(
            q, type_in=search_type, source_in=source_in, return_err=True
        ):
            if result.err_msg:
                hint_msgs.append(
                    t(
                        "track-search-error",
                        providerName=result.source,
                        errorMessage=result.err_msg,
                    )
                )
                view.hint.show_msg("\n".join(hint_msgs))
                continue

            table_container = TableContainer(app, view.accordion)
            table_container.layout().setContentsMargins(0, 0, 0, 0)

            # HACK: set fixed row for tables.
            # pylint: disable=protected-access
            for table in table_container._tables:
                assert isinstance(table, QAbstractItemView)
                delegate = table.itemDelegate()
                if isinstance(delegate, ImgCardListDelegate):
                    table.set_fixed_row_count(2)
                    delegate.update_settings("card_min_width", 140)
                elif isinstance(table, SongsTableView):
                    table.set_fixed_row_count(8)
                    table.set_row_height(table.verticalHeader().defaultSectionSize())

            renderer = SearchResultRenderer(q, tab_index, source_in=source_in)
            await table_container.set_renderer(renderer)
            _, search_type, attrname, show_handler = renderer.tabs[tab_index]
            objects = getattr(result, attrname) or []
            if not objects:  # Result is empty.
                hint_msgs.append(
                    t("track-search-result-empty", providerName=result.source)
                )
                view.hint.show_msg("\n".join(hint_msgs))
                continue

            succeed += 1
            if search_type is SearchType.so:
                show_handler(  # type: ignore[operator]
                    create_reader(objects), columns_mode=ColumnsMode.playlist
                )
            else:
                show_handler(create_reader(objects))  # type: ignore[operator]
            source = objects[0].source
            provider = app.library.get(source)
            provider_name = provider.name
            if is_first is False:
                table_container.hide()
            view.accordion.add_section(MidHeader(provider_name), table_container, 6, 12)
            renderer.meta_widget.hide()
            renderer.toolbar.hide()
            is_first = False
        time_cost = (datetime.now() - start).total_seconds()
        hint_msgs.pop(0)
        hint_msgs.insert(
            0,
            t("track-search-done", resultCount=succeed, timeCost=time_cost),
        )
        view.hint.show_msg("\n".join(hint_msgs))


class SearchResultRenderer(Renderer, TabBarRendererMixin):
    def __init__(self, q, tab_index, source_in=None):
        self.q = q
        self.tab_index = tab_index
        self.source_in = source_in

        self.tabs = [
            (*Tabs[0], "songs", self.show_songs),
            (*Tabs[1], "albums", self.show_albums),
            (*Tabs[2], "artists", self.show_artists),
            (*Tabs[3], "playlists", self.show_playlists),
            (*Tabs[4], "videos", self.show_videos),
        ]

    async def render(self):
        self.render_tab_bar()

    def render_by_tab_index(self, tab_index):
        search_type = self.tabs[tab_index][1]
        self._app.browser.local_storage[KeyType] = search_type.value
        query = {"q": self.q, "type": search_type.value}
        source_in = self._app.browser.local_storage.get(KeySourceIn, None)
        if source_in is not None:
            query["source_in"] = source_in
        self._app.browser.goto(page="/search", query=query)



from PyQt6.QtCore import Qt, QRectF
from PyQt6.QtWidgets import QWidget, QHBoxLayout, QLabel
from PyQt6.QtGui import QTextOption, QColor, QPalette, QPainter

from feeluown.gui.helpers import SOLARIZED_COLORS


class StatusLineItem:
    """Status bar component"""

    def __init__(self, name, widget):
        self.name = name
        self.widget = widget

    def __eq__(self, other):
        return (
            isinstance(other, StatusLineItem)
            and other.name == self.name
            and other.widget is self.widget
        )


class StatuslineLabel(QLabel):
    """
    This is a base class;
    the specific design hasn’t been fully figured out yet,
    and it’s not recommended for widespread use.

    The general idea is as follows: the Label consists of two parts, the middle and
    the notification parts (see Chrome’s extension icon for reference).
    The base class will define the sizes of these two parts and provide some
    predefined drawing functions for subclasses to use.
    """

    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._app = app

        # TODO: The height of the forward and back buttons in the top navigation bar
        # is also 30, and it should be passed in externally later
        self._width = self._height = 30
        self._inner_width = self._inner_height = 18
        self._status_width = self._status_height = 13
        self._status_font_size = 8
        self._status_color = "red"
        self._pressed = False

        self.setFixedSize(self._width, self._height)

    def mousePressEvent(self, e):
        super().mousePressEvent(e)
        self._pressed = True
        self.update()

    def mouseReleaseEvent(self, e):
        super().mousePressEvent(e)
        self._pressed = False
        self.update()

    def paintEvent(self, e):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        self.drawBg(painter)

        painter.save()
        x_delta = (self._width - self._inner_height) // 2
        y_delta = (self._height - self._inner_height) // 2
        painter.translate(x_delta, y_delta)
        self.drawInner(painter)
        painter.restore()

        self.drawStatus(painter)

    def drawInner(self, painter):
        pass

    def drawBg(self, painter):
        if not self._pressed:
            return
        radius = self._width // 2
        painter.save()
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(QColor(self._app.theme_mgr.get_pressed_color()))
        painter.drawRoundedRect(self.rect(), radius, radius)
        painter.restore()

    def drawStatus(self, painter):
        if not self.text():
            return

        painter.save()
        font = self.font()
        font.setPixelSize(self._status_font_size)
        painter.setFont(font)
        pen = painter.pen()
        border_color = self.palette().color(QPalette.ColorRole.Window)
        pen.setColor(QColor(border_color))
        painter.setPen(pen)
        w = h = self._width // 2 - 2
        painter.translate(self._width - w, self._width - h)
        text_rect = QRectF(0, 0, w, h)
        text_bg_color = QColor(SOLARIZED_COLORS[self._status_color])
        text_bg_color.setAlpha(225)
        painter.setBrush(text_bg_color)
        radius = 4
        painter.drawRoundedRect(text_rect, radius, radius)
        painter.drawText(
            text_rect, self.text(), QTextOption(Qt.AlignmentFlag.AlignCenter)
        )
        painter.restore()


class StatusLine(QWidget):
    """Status bar (similar to Emacs/Vim status line)"""

    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app
        self._items = []
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(5)

    def get_item(self, name):
        for item in self._items:
            if item.name == name:
                return item

    def add_item(self, item):
        """Add component"""
        if item not in self._items:
            self._items.append(item)
            self._layout.addWidget(item.widget)

    def remove_item(self, item):
        """Remove component

        :param item: A StatusLineItem object, or an item name
        """
        pass



from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    runtime_checkable,
    Protocol,
    Dict,
    Optional,
    List,
    Callable,
    Any,
)

from PyQt6.QtCore import pyqtSignal, QObject
from PyQt6.QtWidgets import QMenu

from feeluown.library import ProviderV2
from feeluown.gui.widgets.provider import ProvidersModel
from feeluown.utils.dispatch import Signal

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


@dataclass
class NavBtn:
    icon: str
    text: str
    cb: Callable[[], Any]


@runtime_checkable
class UISupportsLoginOrGoHome(Protocol):
    @abstractmethod
    def login_or_go_home(self):
        """This method is called when the avatar is clicked.

        Typically, ProviderUI can implement this method as follows:
        - When no user is logged in, ProviderUI MAY show a login dialog.
        - When a user is logged in, ProviderUI MAY show the homepage.
          At the same time, ProviderUI MAY emit a login success event.
        """


@runtime_checkable
class UISupportsLoginEvent(Protocol):
    """
    When the user is logged in, ProviderUI MAY emit a login success event.
    This allows FeelUOwn to perform certain actions based on this event,
    for example, fetching and show the user's playlist.
    """

    @property
    @abstractmethod
    def login_event(self):
        """
        event:
          0: first login failed
          1: first login ok
          2: re-login ok

        :return: Signal(provider_ui, event)
        """


@runtime_checkable
class UISupportsDiscovery(Protocol):
    @abstractmethod
    def discovery(self): ...


@runtime_checkable
class UISupportsNavBtns(Protocol):
    """
    Provider UI can add its own navigation buttons to the sidebar.
    """

    @abstractmethod
    def list_nav_btns(self) -> List[NavBtn]: ...


@runtime_checkable
class UISupportsContextMenuAddItems(Protocol):
    """
    Provider UI can add custom actions/submenus to the context menu.
    """

    @abstractmethod
    def context_menu_add_items(self, menu: QMenu):
        """
        Add custom actions/submenus to the given menu.
        """


@runtime_checkable
class UISupportsCreatePlaylist(Protocol):
    """
    Provider UI can create playlist.
    """

    @abstractmethod
    def create_playlist(self): ...


class AbstractProviderUi(ABC):
    """Abstract base class for provider ui."""

    def get_colorful_svg(self) -> str:
        """Get provider's svg icon path."""
        return "icons:feeluown.png"

    def register_pages(self, route): ...

    @property
    @abstractmethod
    def provider(self) -> ProviderV2: ...


class CurrentProviderUiManager(QObject):
    changed = pyqtSignal([object, object])

    def __init__(self, app: "GuiApp"):
        super().__init__(parent=app)
        self._app = app
        self._current: Optional[AbstractProviderUi] = None

        # Old code uses `ProviderUiItem` to represent a provider.
        # New code should use `AbstractProviderUI` instead.
        self._current_item: Optional[ProviderUiItem] = None

    def get(self):
        return self._current

    def set(self, provider_ui: AbstractProviderUi):
        self._current_item = None
        old = self._current
        self._current = provider_ui
        self.changed.emit(provider_ui, old)

    def get_either(self):
        return self._current or self._current_item

    def get_item(self):
        return self._current_item

    def set_item(self, item: "ProviderUiItem"):
        self._current = None
        self._current_item = item


class ProviderUiManager:
    """
    Note that `*_item` APIs are deprecated, new code should not use them anymore.
    """

    def __init__(self, app: "GuiApp"):
        self._app = app

        self._store: Dict[str, AbstractProviderUi] = {}  # {name: provider_ui}

        self._items: Dict[str, ProviderUiItem] = {}  # name:model mapping
        self.model = ProvidersModel(self._app.library, self._app)

    def register(self, provider_ui: AbstractProviderUi):
        name = provider_ui.provider.meta.identifier
        if name in self._store or name in self._items:
            raise ValueError(f"provider_ui {name} already registered.")
        self._store[name] = provider_ui

        provider_ui.register_pages(self._app.browser.route)

    def get(self, identifier) -> Optional[AbstractProviderUi]:
        return self._store.get(identifier)

    def list_all(self) -> List[AbstractProviderUi]:
        return list(self._store.values())

    ####################################
    # The following are Deprecated APIs.
    ####################################
    def create_item(self, name, text, symbol="♬ ", desc="", colorful_svg=None):
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        provider = self._app.library.get(name)
        return ProviderUiItem(
            name, text, symbol, desc, colorful_svg=colorful_svg, provider=provider
        )

    def get_item(self, name):
        return self._items.get(name)

    def list_items(self):
        return list(self._items.values())

    def add_item(self, uiitem):
        name = uiitem.name
        self.model.assoc(name, uiitem)
        self._items[name] = uiitem
        return True

    def remove_item(self, uiitem):
        if isinstance(uiitem, ProviderUiItem):
            name = uiitem.name
        else:
            name = uiitem
        if name in self._items:
            self.model.remove(name)
            self._items.pop(name)
            return True
        return False


class ProviderUiItem:
    """
    .. versiondeprecated:: 3.8.15
        Use `AbstractProviderUI` instead.
    """

    def __init__(self, name, text, symbol, desc, colorful_svg=None, provider=None):
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        # If needed, you can support a right-click pop-up menu
        self._name = name
        self.text = text
        self.symbol = symbol
        self.desc = desc
        self.colorful_svg = colorful_svg
        self.provider = provider
        self.clicked = Signal()

    @property
    def name(self):
        return self._name



from feeluown.app.gui_app import GuiApp
from feeluown.i18n import t
from feeluown.library import ModelType
from feeluown.utils.aio import run_fn, run_afn
from feeluown.gui.page_containers.table import Renderer, TableContainer
from feeluown.gui.page_containers.scroll_area import ScrollArea
from feeluown.gui.base_renderer import TabBarRendererMixin
from feeluown.library import (
    SupportsCurrentUserDislikeSongsReader,
    SupportsCurrentUserDislikeRemoveSong,
)
from feeluown.utils.reader import create_reader
from .template import render_error_message


async def render(req, **kwargs):
    app: GuiApp = req.ctx["app"]
    tab_index = int(req.query.get("tab_index", 0))
    pvd_ui = app.current_pvd_ui_mgr.get()
    if pvd_ui is None:
        return await render_error_message(app, t("provider-unknown-cannot-view"))

    scroll_area = ScrollArea()
    # it should not use TableContainer.songs_table to show
    # the dislike song list. For example, the TableContainer.songs_table
    # has a fixed context menu, which is not suitable for the dislike song list.
    # TODO: Maybe design a widget to show the dislike list.
    table_container = TableContainer(app, scroll_area)
    scroll_area.setWidget(table_container)
    app.ui.right_panel.set_body(scroll_area)
    renderer = DislikeRenderer(app, tab_index, pvd_ui.provider)
    await table_container.set_renderer(renderer)


class DislikeRenderer(Renderer, TabBarRendererMixin):
    def __init__(self, app, tab_index, provider):
        self._app = app
        self._provider = provider
        self.tab_index = tab_index
        self.tabs = self.default_tabs()[:1]

    async def render(self):
        self.meta_widget.show()
        self.meta_widget.title = t("music-blacklisted")
        self.render_tab_bar()
        await self.render_models()

    async def dislike_remove_song(self, song):
        await run_fn(self._provider.current_user_dislike_remove_song, song)
        # Workaround: re-render to refresh the page
        await self.render()

    async def render_models(self):
        _, mtype, show_handler = self.tabs[self.tab_index]
        res_type = ""
        reader = create_reader([])
        if mtype is ModelType.song:
            if isinstance(self._provider, SupportsCurrentUserDislikeSongsReader):
                reader = await run_fn(
                    self._provider.current_user_dislike_create_songs_rd
                )  # noqa
            else:
                res_type = "dislike"
            if isinstance(self._provider, SupportsCurrentUserDislikeRemoveSong):
                self.songs_table.remove_song_func = lambda song: run_afn(
                    self.dislike_remove_song, song
                )
        else:
            res_type = "unknown"

        if res_type:
            return await render_error_message(
                self._app,
                t(
                    "provider-unsupported",
                    providerName=self._provider.name,
                    resType=res_type,
                ),
            )
        else:
            show_handler(reader)
        # Do not show playall button.
        self.toolbar.hide()

    def render_by_tab_index(self, tab_index):
        self._app.browser.goto(page="/my_dislike", query={"tab_index": tab_index})



from PyQt6.QtCore import pyqtSignal, Qt
from PyQt6.QtGui import QPainter, QPalette
from PyQt6.QtWidgets import QAbstractSlider

from feeluown.i18n import t
from feeluown.gui.drawers import VolumeIconDrawer
from feeluown.gui.helpers import painter_save, darker_or_lighter


class VolumeButton(QAbstractSlider):
    change_volume_needed = pyqtSignal([int])

    def __init__(self, length=30, padding=0.25, parent=None):
        super().__init__(parent=parent)

        self.setToolTip(t("volume-button-tooltip"))

        font = self.font()
        font.setPixelSize(length // 3)
        self.setFont(font)

        self.__pressed = False
        self.__checked = False

        self.setMinimum(0)
        self.setMaximum(100)

        padding = int(length * padding if padding < 1 else padding)
        self.drawer = VolumeIconDrawer(length, padding)
        self.valueChanged.connect(self.change_volume_needed.emit)
        self.setFixedSize(length, length)

    def on_volume_changed(self, value):
        # blockSignals to avoid circular setVolume.
        # https://stackoverflow.com/a/4146392/4302892
        self.blockSignals(True)
        self.setValue(value)
        self.drawer.set_volume(value)
        self.blockSignals(False)
        self.update()

    def paintEvent(self, _) -> None:
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        if self.__checked is True:
            with painter_save(painter):
                painter.setPen(Qt.PenStyle.NoPen)
                color = self.palette().color(QPalette.ColorRole.Window)
                painter.setBrush(darker_or_lighter(color, 120))
                painter.drawEllipse(self.rect())
            painter.drawText(
                self.rect(), Qt.AlignmentFlag.AlignCenter, f"{self.value()}%"
            )
        else:
            self.drawer.draw(painter, self.palette())

    def mousePressEvent(self, e) -> None:
        super().mousePressEvent(e)
        if e.button() == Qt.MouseButton.LeftButton:
            self.__pressed = True

    def mouseReleaseEvent(self, e):
        super().mouseReleaseEvent(e)
        if e.button() == Qt.MouseButton.LeftButton:
            if self.__pressed is True:
                self.__pressed = False
                self.__checked = not self.__checked
                # schedule an update to refresh ASAP.
                self.update()


if __name__ == "__main__":
    from feeluown.gui.debug import simple_layout

    with simple_layout() as layout:
        layout.addWidget(VolumeButton(100))



import logging
from PyQt6.QtWidgets import QSizePolicy, QSplitter, QVBoxLayout

from feeluown.gui.widgets.separator import Separator
from feeluown.gui.widgets.settings import SettingsDialog
from feeluown.gui.widgets.messageline import MessageLine
from feeluown.gui.widgets.mpv_ import MpvOpenGLWidget

from feeluown.gui.uimain.lyric import LyricWindow
from feeluown.gui.uimain.sidebar import LeftPanel
from feeluown.gui.uimain.page_view import RightPanel
from feeluown.gui.uimain.player_bar import TopPanel
from feeluown.gui.uimain.playlist_overlay import PlaylistOverlay
from feeluown.gui.uimain.nowplaying_overlay import NowplayingOverlay

logger = logging.getLogger(__name__)


class Ui:

    def __init__(self, app):
        self._app = app
        # Let the following widgets access ui object during init.
        self._app.ui = self
        self._layout = QVBoxLayout(app)
        self._top_separator = Separator(app)
        self._splitter = QSplitter(app)

        # Create widgets that don't rely on other widgets first.
        self.ai_chat_overlay = None
        try:
            from feeluown.gui.uimain.ai_chat import create_aichat_overlay
        except ImportError as e:
            logger.warning(f"AIChatOverlay is not available: {e}")
        else:
            if self._app.ai is not None:
                self.ai_chat_overlay = create_aichat_overlay(app, parent=app)
                self.ai_chat_overlay.hide()
        self.lyric_window = LyricWindow(self._app)
        self.lyric_window.hide()
        self.playlist_overlay = PlaylistOverlay(app, parent=app)
        self.nowplaying_overlay = NowplayingOverlay(app, parent=app)

        # NOTE: Components named by position should only be used
        # to organize the UI layout, o not add any functional methods to them.
        self._message_line = MessageLine()
        self.top_panel = TopPanel(app, app)
        self.sidebar = self._left_panel_con = LeftPanel(self._app)
        self.left_panel = self._left_panel_con.p
        self.page_view = self.right_panel = RightPanel(self._app, self._splitter)
        self.toolbar = self.bottom_panel = self.right_panel.bottom_panel
        self.mpv_widget = MpvOpenGLWidget(self._app)

        # alias
        self.magicbox = self.bottom_panel.magicbox
        self.player_bar = self.pc_panel = self.top_panel.pc_panel
        self.table_container = self.right_panel.table_container
        # backward compatible, old name is songs_table_container
        self.songs_table_container = self.table_container
        self.songs_table = self.table_container.songs_table
        self.back_btn = self.bottom_panel.back_btn
        self.forward_btn = self.bottom_panel.forward_btn

        self.toolbar.settings_btn.clicked.connect(self._open_settings_dialog)

        self._setup_ui()

    def _setup_ui(self):
        self._app.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )

        self._splitter.setHandleWidth(0)
        self._splitter.addWidget(self._left_panel_con)
        self._splitter.addWidget(self.right_panel)
        self._message_line.hide()
        self.playlist_overlay.hide()
        self.nowplaying_overlay.hide()

        self.right_panel.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )

        self._layout.addWidget(self._splitter)
        self._layout.addWidget(self.mpv_widget)
        self._layout.addWidget(self._message_line)
        self._layout.addWidget(self._top_separator)
        self._layout.addWidget(self.top_panel)

        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.top_panel.layout().setSpacing(0)
        self.top_panel.layout().setContentsMargins(0, 0, 0, 0)

        self._app.resize(960, 600)

    def _open_settings_dialog(self):
        dialog = SettingsDialog(self._app, self._app)
        dialog.exec()

    def toggle_player_bar(self):
        if self.top_panel.isVisible():
            self.top_panel.hide()
            self._top_separator.hide()
        else:
            self.top_panel.show()
            self._top_separator.show()



import logging
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlparse

from feeluown.i18n import t

from requests import ConnectTimeout


if TYPE_CHECKING:
    from feeluown.app import App

logger = logging.getLogger(__name__)


class AlertManager:
    """Monitor app exceptions and send some alerts."""

    def __init__(self):
        # Some alerts handling rely on app and some are not.
        self._app: Optional["App"] = None

    def initialize(self, app: "App"):
        """"""
        self._app = app
        self._app.player.media_loading_failed.connect(
            self.on_media_loading_failed, aioqueue=True
        )

    def on_exception(self, e):
        if isinstance(e, ConnectTimeout):
            if e.request is not None:
                url = e.request.url
                hostname = urlparse(url).hostname
            else:
                hostname = "none"
            self.show_alert("connection-timeout", hostname=hostname)

    def on_media_loading_failed(self, *_):
        assert self._app is not None
        media = self._app.player.current_media
        if media and media.url:
            proxy = media.http_proxy if media.http_proxy else "none"
            hostname = urlparse(media.url).hostname
            self.show_alert('media-loading-failed', hostname=hostname, proxy=proxy)

    def show_alert(self, msg_id: str, **kwargs):
        logger.warning(t(msg_id, locale="en-US", **kwargs))
        self._app.show_msg(t(msg_id, **kwargs), timeout=2000)



from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QLabel, QMenu

from feeluown.utils.aio import run_afn
from feeluown.media import MediaType
from feeluown.i18n import t

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class SongSourceTag(QLabel):
    default_text = t("music-source")

    def __init__(self, app: "GuiApp", font_size=12, parent=None):
        super().__init__(text=SongSourceTag.default_text, parent=parent)
        self._app = app

        font = self.font()
        font.setPixelSize(font_size)
        self.setFont(font)

        self._source = self.default_text
        self._bitrate = 0

        self._app.player.metadata_changed.connect(
            self.on_metadata_changed, aioqueue=True
        )
        self._app.player.audio_bitrate_changed.connect(
            self.on_audio_bitrate_changed, aioqueue=True
        )

    def contextMenuEvent(self, e):
        # pylint: disable=unnecessary-direct-lambda-call
        # FIXME(wuliaotc): Disable the menu when switching provider
        song = self._app.playlist.current_song
        if song is None:
            return

        menu = QMenu()
        submenu = menu.addMenu(t("track-smart-standby"))
        for provider in self._app.library.list():
            pid = provider.identifier
            if pid == song.source:
                continue
            action = submenu.addAction(provider.name)
            action.triggered.connect(
                (lambda x: lambda: run_afn(self._switch_provider, x))(pid)
            )
        menu.exec(e.globalPos())

    def on_audio_bitrate_changed(self, bitrate):
        if bitrate is not None:
            old_bitrate = self._bitrate
            new_bitrate = bitrate // 1000
            # mpv's audio bitrate calculated on the packet level,
            # only update text when the bitrate changes a lot.
            self._bitrate = new_bitrate
            if new_bitrate - old_bitrate > 10:
                self.update_text()

    def on_metadata_changed(self, metadata):
        if not metadata:
            self._source = ""
            self._bitrate = 0
        else:
            self._source = t("track-unknown-source")
            # Fill source name.
            source = metadata.get("source", "")
            if source:
                source_name_map = {
                    p.identifier: p.name for p in self._app.library.list()
                }
                self._source = source_name_map.get(source, self._source)

            # Fill audio bitrate info if available.
            media = self._app.player.current_media
            if media is not None and media.type_ == MediaType.audio:
                props = media.props
                if props.bitrate:
                    self._bitrate = props.bitrate
        self.update_text()

    def update_text(self):
        if self._source:
            text = f"{self._source} • {self._bitrate}kbps"
            self.setText(text)
        else:
            self.setText(self.default_text)

    async def _switch_provider(self, provider_id):
        song = self._app.playlist.current_song
        songs = await self._app.library.a_list_song_standby_v2(
            song, source_in=[provider_id]
        )
        if songs:
            standby, media = songs[0]
            assert standby != song
            self._app.show_msg(t("track-fallback-to-standby", standby=standby))
            self._app.playlist.set_current_song_with_media(standby, media)
            self._app.playlist.remove(song)
        else:
            self._app.show_msg(t("track-fallback-failed", providerName=provider_id))



from feeluown.library import BaseModel

from .typename import attach_typename, get_type_by_name, model_cls_list
from .base import Serializer, SerializerMeta, DeserializerError


class PythonDeserializer:
    _mapping = {}

    def deserialize(self, obj):
        if isinstance(obj, dict):
            typename = obj['__type__']
            deserializer_cls = self.get_deserializer_cls(typename)
            return deserializer_cls().deserialize(obj)
        elif isinstance(obj, list):
            return [self.deserialize(each) for each in obj]
        if isinstance(obj, (str, int, float, type(None))):
            return obj
        raise DeserializerError(f'no deserializer for type:{type(obj)}')

    @classmethod
    def get_deserializer_cls(cls, typename):
        clz = get_type_by_name(typename)
        if clz is None:
            raise DeserializerError(f'no deserializer for type:{typename}')
        for obj_clz, deserializer_cls in cls._mapping.items():
            if obj_clz == clz:
                return deserializer_cls
        raise DeserializerError(f'no deserializer for type:{clz}')


class ModelDeserializer(PythonDeserializer, metaclass=SerializerMeta):
    class Meta:
        types = model_cls_list

    def deserialize(self, obj):
        model_cls = get_type_by_name(obj['__type__'])
        return model_cls.model_validate(obj)


class PythonSerializer(Serializer):
    _mapping = {}

    def __init__(self, **options):
        super().__init__(**options)

    def serialize_items(self, items):
        json_ = {}
        for key, value in items:
            serializer_cls = PythonSerializer.get_serializer_cls(value)
            serializer = serializer_cls()
            json_[key] = serializer.serialize(value)
        return json_

    @attach_typename
    def serialize(self, obj):
        return super().serialize(obj)


class ModelSerializer(PythonSerializer, metaclass=SerializerMeta):
    class Meta:
        types = (BaseModel, )

    def serialize(self, model: BaseModel):
        return model.model_dump()


#######################
# container serializers
#######################


class ListSerializer(PythonSerializer, metaclass=SerializerMeta):
    class Meta:
        types = (list, )

    def serialize(self, list_):
        if not list_:
            return []
        result = []
        for item in list_:
            serializer_cls = PythonSerializer.get_serializer_cls(item)
            serializer = serializer_cls()
            result.append(serializer.serialize(item))
        return result

    def serialize_search_result_list(self, list_):
        from .objs import SearchPythonSerializer
        serializer = SearchPythonSerializer()
        return [serializer.serialize(model) for model in list_]


####################
# object serializers
####################


class SimpleTypeSerializer(PythonSerializer, metaclass=SerializerMeta):
    class Meta:
        types = (str, int, float, type(None), dict)

    def serialize(self, object):
        return object



import logging

from PyQt6.QtCore import (
    pyqtSignal,
    Qt,
    QModelIndex,
)
from PyQt6.QtWidgets import QAbstractItemView, QMenu

from feeluown.i18n import t
from feeluown.library import SupportsPlaylistAddSong
from feeluown.utils.aio import run_afn, run_fn
from .textlist import TextlistModel, TextlistView


logger = logging.getLogger(__name__)


class PlaylistsModel(TextlistModel):
    def __init__(self, parent):
        super().__init__(parent)
        self._playlists = []
        self._fav_playlists = []

    @property
    def items(self):
        return self._playlists + self._fav_playlists

    def add(self, playlist, is_fav=False):
        if is_fav:
            start = len(self._playlists) + len(self._fav_playlists)
            playlists = self._fav_playlists
        else:
            start = len(self._playlists)
            playlists = self._playlists

        if isinstance(playlist, list):
            _playlists = playlist
        else:
            _playlists = [playlist]
        end = start + len(_playlists)
        self.beginInsertRows(QModelIndex(), start, end)
        playlists.extend(_playlists)
        self.endInsertRows()

    def remove(self, playlist):
        for i, playlist_ in enumerate(self._playlists):
            if playlist_ == playlist:
                self.beginRemoveRows(QModelIndex(), i, i + 1)
                self._playlists.remove(playlist)
                self.endRemoveRows()
                break

        for i, playlist_ in enumerate(self._fav_playlists):
            if playlist_ == playlist:
                start = i + len(self._playlists)
                end = start + 1
                self.beginRemoveRows(QModelIndex(), start, end)
                self._fav_playlists.remove(playlist)
                self.endRemoveRows()
                break

    def clear(self):
        total_length = len(self.items)
        self.beginRemoveRows(QModelIndex(), 0, total_length - 1)
        self._playlists = []
        self._fav_playlists = []
        self.endRemoveRows()

    def flags(self, index):
        if not index.isValid():
            return 0
        flags = Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled
        if index.row() < len(self._playlists):
            flags |= Qt.ItemFlag.ItemIsDropEnabled
        return flags

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        row = index.row()
        playlist = self.items[row]
        if role == Qt.ItemDataRole.DisplayRole:
            if row < len(self._playlists):
                flag = "♬ "
            else:
                flag = "★ "
            return flag + playlist.name
        return super().data(index, role)


class PlaylistsView(TextlistView):
    """Playlist list view

    This view will display all items; in theory there will be no scroll bar and
    it does not accept scroll events

    .. versiondeprecated:: 3.9
    """

    show_playlist = pyqtSignal([object])
    remove_playlist = pyqtSignal([object])

    def __init__(self, parent):
        super().__init__(parent)

        self.setDragDropMode(QAbstractItemView.DragDropMode.DropOnly)
        self.clicked.connect(self._on_clicked)

    def _on_clicked(self, index):
        playlist = index.data(role=Qt.ItemDataRole.UserRole)
        self.show_playlist.emit(playlist)

    def contextMenuEvent(self, event):
        indexes = self.selectionModel().selectedIndexes()
        if len(indexes) != 1:
            return

        playlist = self.model().data(indexes[0], Qt.ItemDataRole.UserRole)
        menu = QMenu()
        action = menu.addAction(t("track-list-remove"))
        action.triggered.connect(lambda: self.remove_playlist.emit(playlist))
        menu.exec(event.globalPos())

    def dropEvent(self, e):
        mimedata = e.mimeData()
        song = mimedata.model
        index = self.indexAt(e.position().toPoint())
        playlist = index.data(Qt.ItemDataRole.UserRole)
        self._results[index.row] = (index, None)
        self.viewport().update()

        async def do():
            is_success = False
            app = self.parent().parent()._app  # type: ignore[attr-defined]
            try:
                provider = app.library.get(playlist.source)
                if isinstance(provider, SupportsPlaylistAddSong):
                    is_success = await run_fn(
                        provider.playlist_add_song, playlist, song
                    )
            except Exception:  # noqa, to avoid crash.
                logger.exception("add song to playlist failed")
                is_success = False
            app.show_msg(
                t("playlist-add-track", status="succ" if is_success else "fail")
            )
            self._results[index.row] = (index, is_success)
            self.viewport().update()
            self._result_timer.start(2000)

        run_afn(do)
        e.accept()

    def dragMoveEvent(self, e):
        mimedata = e.mimeData()
        song = mimedata.model
        index = self.indexAt(e.position().toPoint())
        playlist = index.data(Qt.ItemDataRole.UserRole)
        if song.source == playlist.source:
            e.accept()
            return
        e.ignore()

    def dragEnterEvent(self, e):
        mimedata = e.mimeData()
        if mimedata.hasFormat("fuo-model/x-song"):
            e.accept()
            return
        e.ignore()



from feeluown.utils.dispatch import Signal
from feeluown.gui.widgets.my_music import MyMusicModel


class MyMusicItem:
    def __init__(self, text):
        self.text = text
        self.clicked = Signal()


class MyMusicUiManager:
    """

    .. note::

    Currently, we use an array data structure to store items,
    only providing add_item and clear methods.
    We want the items in MyMusic to remain associated with the provider.
    The provider is the context of MyMusic.

    And the Provider is a relatively higher-level object;
    we will provide get_item for more fine-grained control.
    """

    def __init__(self, app):
        self._app = app
        self._items = []
        self.model = MyMusicModel(app)

    @classmethod
    def create_item(cls, text):
        return MyMusicItem(text)

    def add_item(self, item):
        self.model.add(item)
        self._items.append(item)

    def clear(self):
        self._items.clear()
        self.model.clear()



import logging
from typing import Optional, TYPE_CHECKING

from PyQt6.QtGui import QGuiApplication

from feeluown.i18n import t
from feeluown.excs import ProviderIOError
from feeluown.utils.aio import run_fn, run_afn
from feeluown.player import SongRadio
from feeluown.library import (
    SongModel,
    VideoModel,
    BriefSongModel,
    SupportsPlaylistAddSong,
    SupportsCurrentUserListPlaylists,
    SupportsCurrentUser,
)

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


MV_BTN_TEXT = t("play-track-movie")


class SongMenuInitializer:
    def __init__(self, app: "GuiApp", song):
        """
        :type app: feeluown.app.App
        """
        self._app = app
        self._song = song
        self._fetching_artists = False
        self._fetching_mv = False
        self._fetching_user_playlists = False

    def apply(self, menu):
        app = self._app
        song = self._song

        def enter_song_radio(song):
            radio = SongRadio.create(app, song)
            app.fm.activate(radio.fetch_songs_func, reset=False)
            if app.playlist.current_song != song:
                app.playlist.clear()
                self._app.playlist.next()
                self._app.player.resume()
            else:
                for song_ in app.playlist.list().copy():
                    if song_ is not app.playlist.current_song:
                        app.playlist.remove(song_)
                self._app.player.resume()

        def goto_song_explore(song):
            app.browser.goto(model=song, path="/explore")

        async def goto_song_album(song):
            usong: SongModel = await run_fn(self._app.library.song_upgrade, song)
            if usong.album is not None:
                self._app.browser.goto(model=usong.album)
            else:
                self._app.show_msg(t("track-missing-album"))

        menu.hovered.connect(self.on_action_hovered)
        menu.addAction(t("track-search-similar")).triggered.connect(
            lambda: self.show_similar_resource(song)
        )
        artist_menu = menu.addMenu(t("track-show-artist"))
        artist_menu.menuAction().setData({"artists": None, "song": song})
        mv_menu = menu.addMenu(MV_BTN_TEXT)
        mv_menu.menuAction().setData({"mvs": None, "song": song})

        menu.addAction(t("track-show-album")).triggered.connect(
            lambda: run_afn(goto_song_album, song)
        )
        menu.addAction(t("track-enter-radio")).triggered.connect(
            lambda: enter_song_radio(song)
        )
        menu.addAction(t("track-show-detail")).triggered.connect(
            lambda: goto_song_explore(song)
        )

        menu.addSeparator()
        playlist_add_menu = menu.addMenu(t("track-playlist-add"))

        playlist_add_menu.menuAction().setData(
            {"menu_id": "playlist_add", "playlists": None, "song": song}
        )

        ai_menu = menu.addMenu("AI")
        ai_menu.addAction(t("menu-ai-copy-prompt")).triggered.connect(
            lambda: self.copy_ai_prompt(song)
        )

    def copy_ai_prompt(self, song):
        song_title: str = song.title
        song_artists: str = song.artists_name
        prompt = t("menu-ai-prompt", songTitle=song_title, songArtists=song_artists)
        QGuiApplication.clipboard().setText(prompt)
        self._app.show_msg(t("menu-ai-copy-prompt-succeed"))

    def show_similar_resource(self, song):
        from feeluown.gui.components.search import create_search_result_view

        view = create_search_result_view(self._app, song)
        view.show()

    def on_action_hovered(self, action):
        """
        Fetch song.artists when artists_action is hovered. If it is
        already fetched, ignore.
        """
        data = action.data()
        if data is None:  # submenu action
            return
        if "artists" in data:
            self._hover_artists(action, data)
        elif "mvs" in data:
            self._hover_mv(action, data)
        elif "menu_id" in data and data["menu_id"] == "playlist_add":
            if self._fetching_user_playlists is False:
                self._fetching_user_playlists = True
                self._app.task_mgr.run_afn_preemptive(
                    self._hover_playlist_add, action, data
                )

    async def _hover_playlist_add(self, action, data):
        # pylint: disable=unnecessary-direct-lambda-call
        logger.info("playlist-add action is hovered")
        if data["playlists"] is not None:
            return

        song: BriefSongModel = data["song"]
        provider = self._app.library.get(song.source)

        async def add2p(provider: SupportsPlaylistAddSong, playlist, song):
            try:
                ok = await run_fn(provider.playlist_add_song, playlist, song)
            except ProviderIOError as e:
                logger.error(f"add song to playlist failed {e}")
                ok = False
            if ok:
                self._app.show_msg(
                    t("track-playlist-add-succ", playlistName=playlist.name)
                )
            else:
                self._app.show_msg(
                    t("track-playlist-add-fail", playlistName=playlist.name)
                )

        if (
            isinstance(provider, SupportsPlaylistAddSong)
            and isinstance(provider, SupportsCurrentUser)
            and isinstance(provider, SupportsCurrentUserListPlaylists)
            and provider.has_current_user()
        ):
            try:
                playlists = await run_fn(provider.current_user_list_playlists)
            except ProviderIOError as e:
                logger.error(f"fetch user playlists failed {e}")
                playlists = []
            data["playlists"] = playlists
            action.setData(data)
            action.setEnabled(len(playlists) > 0)
            for playlist in playlists:
                try:
                    action_ = action.menu().addAction(playlist.name)
                    action_.triggered.connect(
                        (lambda p: lambda: run_afn(add2p, provider, p, song))(playlist)
                    )
                except RuntimeError:
                    # action may have been deleted.
                    return
        else:
            action.setEnabled(False)
        self._fetching_user_playlists = False

    def _hover_mv(self, action, data):
        def mv_fetched_cb(future):
            self._fetching_mv = False
            try:
                mv: Optional[VideoModel] = future.result()
            except ProviderIOError as e:
                logger.error(f"fetch song mv failed {e}")
                mv = None
            if mv is not None:
                try:
                    mv_action = action.menu().addAction(mv.title)
                    mv_action.triggered.connect(
                        lambda: self._app.playlist.play_model(mv)
                    )
                except RuntimeError:
                    # action may have been deleted.
                    return
            if mv is not None:
                data["mvs"] = [mv]
                action.setText(MV_BTN_TEXT)
            else:
                data["mvs"] = []
                action.setText(t("track-missing-movie"))
                action.setDisabled(True)
            action.setData(data)

        # artists value has not been fetched
        if data["mvs"] is None and self._fetching_mv is False:
            logger.debug("fetch song.mv for actions")
            song = data["song"]
            self._fetching_mv = True
            task = run_fn(self._app.library.song_get_mv, song)
            task.add_done_callback(mv_fetched_cb)

    def _hover_artists(self, action, data):
        # pylint: disable=unnecessary-direct-lambda-call

        def artists_fetched_cb(future):
            self._fetching_artists = False
            artists = future.result()  # ignore the potential exception
            if artists:
                for artist in artists:
                    try:
                        artist_action = action.menu().addAction(artist.name)
                        # create a closure to bind variable artist
                        artist_action.triggered.connect(
                            (lambda x: lambda: self._app.browser.goto(model=x))(artist)
                        )
                    except RuntimeError:
                        # action may have been deleted.
                        return
            data["artists"] = artists or []
            action.setData(data)

        # artists value has not been fetched
        if data["artists"] is None and self._fetching_artists is False:
            logger.debug("fetch song.artists for actions")
            song = data["song"]
            self._fetching_artists = True
            task = run_fn(lambda: self._app.library.song_upgrade(song).artists)
            task.add_done_callback(artists_fetched_cb)



from Foundation import NSRunLoop, NSMutableDictionary, NSObject
from MediaPlayer import MPRemoteCommandCenter, MPNowPlayingInfoCenter
from MediaPlayer import (
    MPMediaItemPropertyTitle, MPMediaItemPropertyArtist,
    MPMusicPlaybackState, MPMusicPlaybackStatePlaying, MPMusicPlaybackStatePaused,
    MPNowPlayingInfoPropertyElapsedPlaybackTime, MPMediaItemPropertyPlaybackDuration
)


class NowPlaying:
    def __init__(self):
        self.cmd_center = MPRemoteCommandCenter.sharedCommandCenter()
        self.info_center = MPNowPlayingInfoCenter.defaultCenter()

        cmds = [
            self.cmd_center.togglePlayPauseCommand(),
            self.cmd_center.playCommand(),
            self.cmd_center.pauseCommand(),
        ]

        for cmd in cmds:
            cmd.addTargetWithHandler_(self._create_handler(cmd))

        self.update_info()

    def update_info(self):
        nowplaying_info = NSMutableDictionary.dictionary()
        nowplaying_info[MPMediaItemPropertyTitle] = "title"
        nowplaying_info[MPMediaItemPropertyArtist] = "artist"
        nowplaying_info[MPNowPlayingInfoPropertyElapsedPlaybackTime] = 0
        nowplaying_info[MPMediaItemPropertyPlaybackDuration] = 100
        self.info_center.setNowPlayingInfo_(nowplaying_info)
        self.info_center.setPlaybackState_(MPMusicPlaybackStatePlaying)


    def _create_handler(self, cmd):

        def handle(event):
            if event.command() == self.cmd_center.pauseCommand():
                self.info_center.setPlaybackState_(MPMusicPlaybackStatePaused)
            elif event.command() == self.cmd_center.playCommand():
                self.info_center.setPlaybackState_(MPMusicPlaybackStatePlaying)
            return 0

        return handle


def runloop():
    """
    HELP: This function can't be called in non-main thread.
    """
    nowplaying = NowPlaying()
    NSRunLoop.mainRunLoop().run()


runloop()



from typing import TYPE_CHECKING

from PyQt6.QtGui import QPainter, QPolygonF, QColor, QTextOption
from PyQt6.QtCore import QPointF, QRectF, Qt

from feeluown.player import PlaylistMode
from feeluown.gui.widgets import SelfPaintAbstractSquareButton
from feeluown.gui.helpers import SOLARIZED_COLORS
from feeluown.i18n import t

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class PlaylistButton(SelfPaintAbstractSquareButton):
    def __init__(self, app: "GuiApp", *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._app = app

        self.setToolTip(t("playlist-show"))
        # PlaylistButton show FM text on the button when FM mode is actiavted,
        # so update the button when mode is changed.
        self._app.playlist.mode_changed.connect(lambda *args: self.update(), weak=False)
        self.clicked.connect(self.raise_playlist_view)

    def raise_playlist_view(self, *_):
        playlist_overlay = self._app.ui.playlist_overlay

        if not playlist_overlay.isVisible():
            width = max(self._app.width() // 4, 330)
            x = self._app.width() - width
            height = self._app.height()
            playlist_overlay.setGeometry(x, 0, width, height)
            playlist_overlay.show()
            playlist_overlay.setFocus()
            # Put the widget on top.
            playlist_overlay.raise_()

    def paintEvent(self, e):
        # pylint: disable=too-many-locals
        width = height = self.width() - self._padding * 2
        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)
        self.paint_round_bg_when_hover(painter)

        painter.save()
        painter.translate((self.width() - width) // 2, (self.height() - height) // 2)

        line_height = 1.3
        line_margin = (height - 3 * line_height) / 3
        h1 = line_margin
        h2 = h1 + line_margin + line_height
        h3 = h2 + line_margin + line_height

        pen = painter.pen()
        pen.setWidthF(line_height)
        painter.setPen(pen)
        painter.setBrush(pen.color())

        # Draw triangle and first line
        triangle_side_length_half = line_margin * 0.6
        triangle_height = triangle_side_length_half * 1.7
        triangle = QPolygonF(
            [
                QPointF(0, h1 - triangle_side_length_half),
                QPointF(triangle_height, h1),
                QPointF(0, h1 + triangle_side_length_half),
            ]
        )
        painter.drawPolygon(triangle)
        painter.drawLine(
            QPointF(triangle_height + triangle_side_length_half, h1), QPointF(width, h1)
        )

        # Draw second line
        painter.drawLine(QPointF(0, h2), QPointF(width, h2))

        # Draw third line, show FM text if needed
        if self._app.playlist.mode is PlaylistMode.fm:
            painter.drawLine(QPointF(0, h3), QPointF(width // 2, h3))
            painter.pen()
            pen.setColor(QColor(SOLARIZED_COLORS["blue"]))
            painter.setPen(pen)
            font = painter.font()
            rect_h_half = line_margin // 2
            font.setPixelSize(int(rect_h_half * 2))
            painter.setFont(font)
            rect = QRectF(
                width // 2 + rect_h_half,
                h3 - rect_h_half,
                width // 2 - rect_h_half,
                rect_h_half * 2,
            )
            painter.drawText(rect, "FM", QTextOption(Qt.AlignmentFlag.AlignCenter))
        else:
            painter.drawLine(QPointF(0, h3), QPointF(width, h3))

        painter.restore()



from textwrap import indent
# FIXME: maybe we should move `reverse` into serializers package
from .base import Serializer, SerializerMeta, SerializerError
from ._plain_formatter import WideFormatter

from feeluown.library import (
    reverse,
    BaseModel,
    SongModel,
    ArtistModel,
    AlbumModel,
    PlaylistModel,
    UserModel,
    BriefSongModel,
    BriefArtistModel,
    BriefAlbumModel,
    BriefPlaylistModel,
    BriefUserModel,
)

formatter = WideFormatter()
fmt = formatter.format


class ModelSerializerMixin:

    def _get_items(self, model):
        # initialize fields that need to be serialized
        # if as_line option is set, we always use fields_display
        modelcls = type(model)
        fields = [field for field in model.model_fields
                  if field not in BaseModel.model_fields]
        # Include properties.
        pydantic_fields = ("__values__", "fields", "__fields_set__",
                           "model_computed_fields", "model_extra",
                           "model_fields_set")
        fields += [prop for prop in dir(modelcls)
                   if isinstance(getattr(modelcls, prop), property)
                   and prop not in pydantic_fields]
        items = [("provider", model.source),
                 ("identifier", str(model.identifier)),
                 ("uri", reverse(model))]
        for field in fields:
            items.append((field, getattr(model, field)))
        return items


class SongSerializerMixin:
    class Meta:
        types = (SongModel, BriefSongModel)
        # since url can be too long, we put it at last
        fields = ('title', 'duration', 'album', 'artists')
        line_fmt = '{uri:{uri_length}}\t# {title:_18} - {artists_name:_20}'


class ArtistSerializerMixin:
    class Meta:
        types = (ArtistModel, BriefArtistModel)
        fields = ('name', 'songs')
        line_fmt = '{uri:{uri_length}}\t# {name:_40}'


class AlbumSerializerMixin:
    class Meta:
        types = (AlbumModel, BriefAlbumModel)
        fields = ('name', 'artists', 'songs')
        line_fmt = '{uri:{uri_length}}\t# {name:_18} - {artists_name:_20}'


class PlaylistSerializerMixin:
    class Meta:
        types = (PlaylistModel, BriefPlaylistModel)
        fields = ('name',)
        line_fmt = '{uri:{uri_length}}\t# {name:_40}'


class UserSerializerMixin:
    class Meta:
        types = (UserModel, BriefUserModel)
        fields = ('name', 'playlists')
        line_fmt = '{uri:{uri_length}}\t# {name:_40}'


class PlainSerializer(Serializer):
    """PlainSerializer base class"""
    _mapping = {}

    def __init__(self, **options):
        super().__init__(**options)
        self.opt_level = options.get('level', 0)

    def serialize_items(self, items):
        key_length = max(len(key) for key, _ in items)
        normal_field_tpl = '{key:>%d}:  {value}' % (key_length + 1)
        list_field_tpl = '{key:>%d}::' % (key_length)
        text_list = []
        level = self.opt_level + 1
        for field, value in items:
            if isinstance(value, list):
                # append key
                text = fmt(list_field_tpl, key=field)
                text_list.append(text)
                # append value with indent
                if value:
                    serialize_cls = self.get_serializer_cls(value)
                    serializer = serialize_cls(fetch=False, level=level)
                    value_text = serializer.serialize(value)
                    text_list.append(indent(value_text, ' ' * (key_length + 4)))
            else:
                serialize_cls = self.get_serializer_cls(value)
                # field value should be formatted as one line
                serializer = serialize_cls(fetch=False, level=level)
                value_text = serializer.serialize(value)
                text = fmt(normal_field_tpl, key=field, value=value_text)
                text_list.append(text)
        return '\n'.join(text_list)


class ModelSerializer(PlainSerializer, ModelSerializerMixin):

    def __init__(self, **options):
        # do some validation
        if all([options.get('as_line') is False,
                options.get('brief') is False,
                options.get('fetch') is False]):
            raise SerializerError(
                "as_line, brief, fetch can't be false at same time")
        if options.get('as_line') is True and options.get('brief') is False:
            raise SerializerError(
                "brief can't be False when as_line is True")
        if options.get('brief') is False and options.get('fetch') is False:
            raise SerializerError(
                "fetch can't be False when brief is False")

        super().__init__(**options)
        self.opt_as_line = options.get('as_line', False)
        self.opt_brief = options.get('brief', True)
        self.opt_fetch = options.get('fetch', not self.opt_brief)
        self.opt_uri_length = options.get('uri_length', '')

    def serialize(self, model):
        """serialize a model"""
        items = self._get_items(model)
        # if level > 0, we think this should be formatted as one line,
        # subclass can override this behavious by overriding serialize method
        if self.opt_as_line or self.opt_level > 0:
            return fmt(self._line_fmt, uri_length=self.opt_uri_length, **dict(items))
        return self.serialize_items(items)


class ListSerializer(PlainSerializer, metaclass=SerializerMeta):
    """
    Theoretically, we should have ListTypeSerializer for each object type,
    but many of these object have exact same serialization logic.

    SearchModel is an exception.
    """

    class Meta:
        types = (list,)

    def serialize(self, list_):
        from .objs import SearchPlainSerializer

        if not list_:
            return ''
        item0 = list_[0]
        serializer_cls = PlainSerializer.get_serializer_cls(item0)
        level = self.opt_level + 1
        if issubclass(serializer_cls, ModelSerializer):
            if issubclass(serializer_cls, SearchPlainSerializer):
                return self.serialize_search_result_list(list_)
            uri_length = max(len(reverse(item)) for item in list_)
            serializer = serializer_cls(fetch=False, level=level,
                                        uri_length=uri_length, )
        else:
            # list item should be formatted as one line
            serializer = serializer_cls(fetch=False, level=level)
        text_list = []
        for item in list_:
            serialized_item = serializer.serialize(item)
            if serialized_item:
                text_list.append(serialized_item)

        return '\n'.join(text_list)

    def serialize_search_result_list(self, list_):
        from .objs import SearchPlainSerializer

        serializer = SearchPlainSerializer(
            level=self.opt_level + 1,
            fetch=False,
            brief=True,
            as_line=False
        )
        # calculate max uri_length
        max_uri_length = 0
        for model in list_:
            max_uri_length = max(serializer.calc_max_uri_length(model),
                                 max_uri_length)
        serializer.opt_uri_length = max_uri_length
        text_list = [serializer.serialize(model) for model in list_]
        return '\n'.join(text_list)


class SongSerializer(ModelSerializer, SongSerializerMixin,
                     metaclass=SerializerMeta):
    pass


class ArtistSerializer(ModelSerializer, ArtistSerializerMixin,
                       metaclass=SerializerMeta):
    pass


class AlbumSerializer(ModelSerializer, AlbumSerializerMixin,
                      metaclass=SerializerMeta):
    pass


class PlaylistSerializer(ModelSerializer, PlaylistSerializerMixin,
                         metaclass=SerializerMeta):
    pass


class UserSerializer(ModelSerializer, UserSerializerMixin,
                     metaclass=SerializerMeta):
    pass


####################
# object serializers
####################


class SimpleTypeSerializer(PlainSerializer, metaclass=SerializerMeta):
    class Meta:
        types = (str, int, float, type(None), bool)

    def serialize(self, object):
        if object is None:
            return 'null'
        elif object is True:
            return 'true'
        elif object is False:
            return 'false'
        return str(object)



from PyQt6.QtWidgets import QFrame


stylesheet = """
QFrame[frameShape="4"],
QFrame[frameShape="5"]
{{
    border: none;
    background: {};
}}
"""


class Separator(QFrame):
    def __init__(self, app, orientation="horizontal"):
        super().__init__(parent=None)

        self._app = app

        if orientation == "horizontal":
            self.setFrameShape(QFrame.Shape.HLine)
        else:
            self.setFrameShape(QFrame.Shape.VLine)

        if self._app.theme_mgr.theme == "dark":
            self.setStyleSheet(stylesheet.format("#232323"))
            self.setMaximumHeight(1)
        else:
            self.setFrameShadow(QFrame.Shadow.Sunken)



from typing import Optional

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QSlider, QAbstractSlider

from feeluown.player import State
from feeluown.gui.helpers import IS_MACOS
from feeluown.i18n import t


class DraggingContext:
    def __init__(self):
        self.is_media_changed = False


class ProgressSlider(QSlider):
    def __init__(self, app, parent=None):
        super().__init__(parent)

        self._app = app
        self._dragging_ctx: Optional[DraggingContext] = None

        self.setToolTip(t("drag-to-seek-progress"))
        self.setRange(0, 0)  # User can't drag the slider control when range is empty.
        self.setOrientation(Qt.Orientation.Horizontal)

        self.sliderPressed.connect(self.on_pressed)
        self.sliderReleased.connect(self.on_released)
        self.actionTriggered.connect(self.on_action_triggered)

        self._app.player.duration_changed.connect(self.update_total, aioqueue=True)
        self._app.player_pos_per300ms.changed.connect(self.update_progress)
        self._app.player.media_changed.connect(self.on_media_changed)

        if IS_MACOS:
            self.setMinimumHeight(25)  # otherwise, the handler can't show properly

    def update_total(self, s):
        s = s or 0
        self.setRange(0, int(s))

    def update_progress(self, s):
        # Only update progress when it is not dragging.
        if not self.is_dragging:
            s = s or 0
            self.setValue(int(s))

    @property
    def is_dragging(self):
        return self._dragging_ctx is not None

    def on_pressed(self):
        self._dragging_ctx = DraggingContext()
        if self._app.player.state is State.playing:
            self._app.player.pause()

    def on_released(self):
        # Only set position if the player has a valid media and the media is changed.
        assert self._dragging_ctx is not None
        if not self._dragging_ctx.is_media_changed:
            self.maybe_update_player_position(self.value())
        self._dragging_ctx = None

        # Update progress after dragging.
        self.update_progress(self._app.player.position)

    def on_media_changed(self, media):
        if self._dragging_ctx is not None:
            self._dragging_ctx.is_media_changed = True

    def on_action_triggered(self, action):
        # SliderAction.SliderMove is handled seperately. Just ignore it.
        if action not in (
            QAbstractSlider.SliderAction.SliderNoAction,
            QAbstractSlider.SliderAction.SliderMove,
        ):
            slider_position = self.sliderPosition()
            self.maybe_update_player_position(slider_position)

    def maybe_update_player_position(self, position):
        if self._app.player.current_media:
            self._app.player.position = position
            # No matter if the player is paused/stopped before dragging,
            # we resume the player. Bilibili web player did the same.
            self._app.player.resume()



import warnings
from typing import Optional

from PyQt6.QtCore import QSize
from PyQt6.QtGui import QPainter, QImage
from PyQt6.QtWidgets import QLabel, QSizePolicy, QMenu

from feeluown.i18n import t
from feeluown.library import SupportsImgUrlToMedia
from feeluown.media import Media, MediaType
from feeluown.gui.drawers import PixmapDrawer
from feeluown.gui.image import open_image


class CoverLabel(QLabel):
    def __init__(self, parent=None, pixmap=None, radius=3):
        super().__init__(parent=parent)

        self._radius = radius
        self.drawer = PixmapDrawer(None, self, self._radius)
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding
        )

    def show_pixmap(self, pixmap):
        """
        .. versiondeprecated:: 3.8.11
        """
        warnings.warn("You should use show_img", DeprecationWarning)
        self.updateGeometry()
        self.update()  # Schedule a repaint to refresh the UI ASAP.

    def show_img(self, img: Optional[QImage]):
        if not img or img.isNull():
            self.drawer = PixmapDrawer(None, self, self._radius)
        else:
            self.drawer = PixmapDrawer(img, self, self._radius)
        self.updateGeometry()
        self.update()

    def paintEvent(self, e):
        """
        draw pixmap with border radius

        We found two way to draw pixmap with border radius,
        one is as follow, the other way is using bitmap mask,
        but in our practice, the mask way has poor render effects
        """
        painter = QPainter(self)
        self.drawer.draw(painter)

    def contextMenuEvent(self, e):
        if self.drawer.get_img() is None:
            return
        menu = QMenu()
        action = menu.addAction(t("show-original-image"))
        action.triggered.connect(lambda: open_image(self.drawer.get_img()))
        menu.exec(e.globalPos())

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self.updateGeometry()

    def sizeHint(self):
        super_size = super().sizeHint()
        pixmap = self.drawer.get_pixmap()
        if pixmap is None:
            return super_size
        h = (self.width() * pixmap.height()) // pixmap.width()
        # cover label height hint can be as large as possible, since the
        # parent width has been set maximumHeigh
        w = self.width()
        return QSize(w, min(w, h))


class CoverLabelV2(CoverLabel):
    """

    .. versionadded:: 3.7.8
    """

    def __init__(self, app, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)

        self._app = app

    async def show_cover_media(self, cover_media: Optional[Media], cover_uid):
        if cover_media is None:
            self.show_img(None)
            return
        content = await self._app.img_mgr.get(cover_media, cover_uid)
        img = QImage()
        img.loadFromData(content)
        self.show_img(img)

    async def show_cover(self, cover, cover_uid):
        # Backward-compatible API. Prefer show_cover_media when caller can
        # provide a provider-aware Media object.
        if isinstance(cover, Media):
            cover_media = cover
        elif isinstance(cover, str):
            cover_media = Media(cover, MediaType.image)
        else:
            raise TypeError(f"unsupported cover type: {type(cover)!r}")
        await self.show_cover_media(cover_media, cover_uid)

    async def show_cover_with_source(
        self, artwork: str, source: str, cover_uid: str
    ):
        if not artwork:
            self.show_img(None)
            return
        provider = self._app.library.get(source) if source else None
        if isinstance(provider, SupportsImgUrlToMedia):
            cover_media = provider.img_url_to_media(artwork)
        else:
            cover_media = Media(artwork, MediaType.image)
        await self.show_cover_media(cover_media, cover_uid)


if __name__ == "__main__":
    from feeluown.gui.debug import simple_layout

    with simple_layout() as layout:
        label = CoverLabel()
        layout.addWidget(label)
        label.resize(100, 100)
        label.show_img(QImage("/path/to/test.png"))



import re
import sys
from collections import namedtuple, deque
from typing import Deque, Any

from feeluown.server.excs import FuoSyntaxError


# Note: Many of the following regular expressions are
# copied from jinja2/lexer.py and pygments/lexer.py
TOKEN_NAME = sys.intern('name')
TOKEN_FURI = sys.intern('furi')  # fuo uri
TOKEN_STRING = sys.intern('string')
TOKEN_UNQUOTE_STRING = sys.intern('unquote-string')
TOKEN_WHITESPACE = sys.intern('whitespace')
TOKEN_INTEGER = sys.intern('integer')
TOKEN_FLOAT = sys.intern('float')
TOKEN_COMMA = sys.intern('comma')
TOKEN_LBRACKET = sys.intern('lbracket')
TOKEN_RBRACKET = sys.intern('rbracket')
TOKEN_EQ = sys.intern('eq')
TOKEN_REQ_DELIMETER = sys.intern('req_delimeter')
TOKEN_HEREDOC_OP = sys.intern('heredoc_op')
TOKEN_HEREDOC_WORD = sys.intern('heredoc_word')

name_re = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*')
whitespace_re = re.compile(r'\s+')
string_re = re.compile(r"('([^'\\]*(?:\\.[^'\\]*)*)'"
                       r'|"([^"\\]*(?:\\.[^"\\]*)*)")', re.S)
furi_re = re.compile(r'fuo://[/\w+\d+]*')
unquote_string_re = re.compile(r'[\w-]+')
integer_re = re.compile(r'\d+')
float_re = re.compile(r'(?<!\.)\d+\.\d+')
newline_re = re.compile(r'(\r\n|\r|\n)')
comma_re = re.compile(r',')
lbracket_re = re.compile(r'\[')
rbracket_re = re.compile(r'\]')
req_delimeter_re = re.compile(r'#:')
equal_re = re.compile(r'=')
heredoc_op_re = re.compile(r'<<')
heredoc_word_re = re.compile(r'EOF|END')

Token = namedtuple('Token', ['column', 'type_', 'value'])

base_rules = [
    (TOKEN_FURI, furi_re),
    (TOKEN_NAME, name_re),
    (TOKEN_WHITESPACE, whitespace_re),
    (TOKEN_FLOAT, float_re),
    (TOKEN_INTEGER, integer_re),
    (TOKEN_STRING, string_re),
    (TOKEN_UNQUOTE_STRING, unquote_string_re),
    (TOKEN_EQ, equal_re),
]

root_rules = base_rules + [
    (TOKEN_LBRACKET, lbracket_re),
    (TOKEN_REQ_DELIMETER, req_delimeter_re),
    (TOKEN_HEREDOC_OP, heredoc_op_re),
]

bracket_rules = base_rules + [
    (TOKEN_COMMA, comma_re),
    (TOKEN_RBRACKET, rbracket_re),
]

req_rules = base_rules + [
    (TOKEN_COMMA, comma_re),
    (TOKEN_HEREDOC_OP, heredoc_op_re),
]

heredoc_rules = [
    (TOKEN_WHITESPACE, whitespace_re),
    (TOKEN_HEREDOC_WORD, heredoc_word_re),
]

state_rules = {
    'root': root_rules,
    'bracket': bracket_rules,
    'req': req_rules,
    'heredoc': heredoc_rules,
}


def get_state_expect(state):
    if state == 'bracket':
        return ']'
    raise ValueError(f'state: {str(state)} has no need')


class Lexer:
    """fuo protocol request lexical analyzer

    >>> list(token.value for token in Lexer().tokenize('play fuo://local'))
    ['play', 'fuo://local']
    >>> text = "play 'fuo://local/songs1 # 晴天 - 周杰伦'"
    >>> list(token.value for token in Lexer().tokenize(text))
    ['play', 'fuo://local/songs1 # 晴天 - 周杰伦']
    >>> text2 = "search 哈哈 [artist=喵]  #: less"
    >>> list(token.value for token in Lexer().tokenize(text2))
    ['search', '哈哈', '[', 'artist', '=', '喵', ']', '#:', 'less']
    """

    def tokenize(self, source):  # pylint: disable=too-many-arguments,too-many-branches

        def err(msg, column):
            raise FuoSyntaxError(msg, text=source, column=column)

        pos = 0

        # The stack is a data structure commonly used in lexer implementations
        # It is very useful in scenarios such as handling whether parentheses are closed;
        # many lexer implementations add a #pop marker to tokens;
        # this marker is also used in conjunction with the stack.
        state_stack: Deque[str] = deque()

        state = 'root'
        while 1:
            for rule in state_rules[state]:
                type_, regex = rule
                m = regex.match(source, pos)
                if m is None:
                    continue
                value: Any = m.group()
                if type_ == TOKEN_STRING:
                    value = value[1:-1]
                elif type_ == TOKEN_INTEGER:
                    value = int(value)
                elif type_ == TOKEN_FLOAT:
                    value = float(value)
                token = Token(pos, type_, value)

                if type_ == TOKEN_LBRACKET:
                    state_stack.append(state)
                    state = 'bracket'
                elif type_ == TOKEN_REQ_DELIMETER:
                    if state != 'root':
                        expect = get_state_expect(state)
                        err(f"expected token '{expect}' at pos {pos}", pos)
                    state_stack.append(state)
                    state = 'req'
                elif type_ == TOKEN_RBRACKET:
                    if state != 'bracket':
                        err(f"unexpected token ']' at pos {pos}", pos)
                    state = state_stack.pop()
                elif type_ == TOKEN_HEREDOC_OP:
                    if state not in ('root', 'req'):
                        err(f"unexpected token '<<' at pos {pos}", pos)
                    state = 'heredoc'
                if token.type_ != TOKEN_WHITESPACE:
                    yield token
                pos = m.end()
                break
            else:
                if pos == len(source):
                    if state not in ('root', 'req', 'heredoc'):
                        expect = get_state_expect(state)
                        err(f"expected token '{expect}', but end of line reached", pos)
                    else:
                        break
                else:
                    err(f"unexpected token '{source[pos]}' at pos {pos}", pos)



import os
import ctypes
from ctypes import wintypes

if os.name == "nt":
    # Load kernel32.dll
    kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)

    # Define GetUserDefaultLocaleName signature
    # int GetUserDefaultLocaleName(LPWSTR lpLocaleName, int cchLocaleName);
    kernel32.GetUserDefaultLocaleName.argtypes = [wintypes.LPWSTR, ctypes.c_int]
    kernel32.GetUserDefaultLocaleName.restype = ctypes.c_int

# Allocate buffer for locale name
LOCALE_NAME_MAX_LENGTH = 85  # per Windows docs


def user_default_locale() -> str:
    """
    Get user default locale encoded in BCP-47.
    """
    buffer = ctypes.create_unicode_buffer(LOCALE_NAME_MAX_LENGTH)

    # Call the function
    result = kernel32.GetUserDefaultLocaleName(buffer, LOCALE_NAME_MAX_LENGTH)

    if result == 0:
        # Call failed
        error_code = ctypes.get_last_error()
        raise ctypes.WinError(error_code)

    return buffer.value


if __name__ == "__main__":
    print(user_default_locale())



from feeluown.app import App
from feeluown.ai.copilot import Copilot


# FIXME: Other packages should only import things from feeluown.ai
#  They should not import things from feeluown.ai.copilot or other inner moduels.


class AI:
    def __init__(self, app: App):
        self._app = app
        self._copilot = Copilot(self._app)

    def get_copilot(self):
        return self._copilot



from PyQt6.QtCore import Qt, QMetaObject, pyqtSlot
from PyQt6.QtWidgets import QOpenGLWidget, QApplication, QWidget, QHBoxLayout
from PyQt6.QtOpenGL import QGLContext

# HELP: currently, we need import GL module，otherwise it will raise seg fault on Linux(Ubuntu 18.04)
# from OpenGL import GL  # noqa

from feeluown.mpv import MPV, OpenGlCbGetProcAddrFn, MpvRenderContext


def get_proc_addr(_, name):
    glctx = QGLContext.currentContext()
    if glctx is None:
        return 0
    addr = int(glctx.getProcAddress(name.decode('utf-8')))
    return addr


class MpvWidget(QOpenGLWidget):
    def __init__(self, window, parent=None):
        super().__init__(parent=parent)
        self.mpv = MPV(ytdl=True)
        self.ctx = None
        self.get_proc_addr_c = OpenGlCbGetProcAddrFn(get_proc_addr)
        self._window = window

    def initializeGL(self):
        params = {'get_proc_address': self.get_proc_addr_c}
        self.ctx = MpvRenderContext(self.mpv,
                                    'opengl',
                                    opengl_init_params=params)
        self.ctx.update_cb = self.on_update

    def paintGL(self):
        # compatible with HiDPI display
        ratio = self._window.devicePixelRatio()
        w = int(self.width() * ratio)
        h = int(self.height() * ratio)
        opengl_fbo = {'w': w,
                      'h': h,
                      'fbo': self.defaultFramebufferObject()}
        self.ctx.render(flip_y=True, opengl_fbo=opengl_fbo)

    @pyqtSlot()
    def maybe_update(self):
        if self.window().isMinimized():
            self.makeCurrent()
            self.paintGL()
            self.context().swapBuffers(self.context().surface())
            self.doneCurrent()
        else:
            self.update()

    def on_update(self, ctx=None):
        # maybe_update method should run on the thread that creates the OpenGLContext,
        # which in general is the main thread. QMetaObject.invokeMethod can
        # do this trick.
        QMetaObject.invokeMethod(self, 'maybe_update')

    def on_update_fake(self, ctx=None):
        pass

    def closeEvent(self, _):
        # self.makeCurrent()
        # self.mpv.terminate()
        pass


if __name__ == '__main__':
    import locale, time, threading

    app = QApplication([])
    locale.setlocale(locale.LC_NUMERIC, 'C')
    root = QWidget()
    layout = QHBoxLayout(root)
    p1 = QWidget()
    p2 = QWidget()
    l1 = QHBoxLayout(p1)
    l2 = QHBoxLayout(p2)
    layout.addWidget(p1)
    layout.addWidget(p2)
    root.resize(600, 400)
    root.show()

    widget = MpvWidget(root)
    l1.addWidget(widget)
    widget.show()

    url = 'data/test.webm'
    widget.mpv.play(url)

    def reparent():
        time.sleep(1)
        widget.mpv.pause = True
        l2.addWidget(widget)
        widget.mpv.pause = False

    threading.Thread(target=reparent).start()
    app.exec()



from .provider_ui import ProviderUiManager
from .uimodels.my_music import MyMusicUiManager
from .uimodels.playlist import PlaylistUiManager

__all__ = (
    "ProviderUiManager",
    "MyMusicUiManager",
    "PlaylistUiManager",
)



import logging
from typing import Optional, List, Any

from yt_dlp import YoutubeDL, DownloadError

from feeluown.media import Media, VideoAudioManifest

logger = logging.getLogger(__name__)


class Ytdl:

    def __init__(self, rules: Optional[List[Any]] = None):
        self._default_ytdl_opts = {
            'logger': logger,
            'socket_timeout': 2,
            'extractor_retries': 0,  # reduce retry
        }
        self._default_audio_ytdl_opts = {
            # The following two options may be only valid for select_audio API.
            # Remove these two options if needed.
            'format': 'm4a/bestaudio/best',
            **self._default_ytdl_opts,
        }
        self._default_video_ytdl_opts = {
            **self._default_ytdl_opts,
        }
        # For example::
        #    [
        #        {'name': 'match_source',
        #         'pattern': 'xxx',
        #         'http_proxy': 'http://127.0.0.1:7890'},
        #    ]
        # TODO: valid rules
        self._rules = rules or []

    def match_rule(self, url, source=''):
        """
        Currently, only 'match_source' is supported. Maybe support 'match_url'
        in the future.
        """
        matched_rule = None
        if isinstance(self._rules, list) and self._rules:
            for rule in self._rules:
                if rule['name'] == 'match_source' and source == rule['pattern']:
                    matched_rule = rule
        if matched_rule:
            logger.info(f"ytdl rule matched for {url}")
        else:
            logger.info(f"no ytdl rule matched for {url}")
        return matched_rule

    def select_audio(self, url, _: Optional[str] = None, source='') -> Optional[Media]:
        matched_rule = self.match_rule(url, source)
        if matched_rule is None:
            return None
        http_proxy = matched_rule.get('http_proxy')
        ytdl_opts = {}
        ytdl_opts.update(self._default_audio_ytdl_opts)
        ytdl_opts['proxy'] = http_proxy
        ytdl_opts.update(matched_rule.get('ytdl_options', {}))
        with YoutubeDL(ytdl_opts) as inner:
            try:
                info = inner.extract_info(url, download=False)
            except DownloadError:  # noqa
                logger.warning(f"extract_info failed for {url}")
                info = None
            if info:
                media_url = info['url']
                if media_url:
                    # NOTE(cosven): do not set http headers, otherwise it can't play.
                    # Tested with 'https://music.youtube.com/watch?v=vKwowKeEv5w'
                    return Media(
                        media_url,
                        format=info['ext'],
                        bitrate=int(info['abr']),
                        http_proxy=http_proxy
                    )
            return None

    def select_video(self, url, _: Optional[str] = None, source='') -> Optional[Media]:
        matched_rule = self.match_rule(url, source)
        if matched_rule is None:
            return None
        http_proxy = matched_rule.get('http_proxy')
        ytdl_opts = {}
        ytdl_opts.update(self._default_video_ytdl_opts)
        ytdl_opts['proxy'] = http_proxy
        ytdl_opts.update(matched_rule.get('ytdl_options', {}))

        audio_candidates = []  # [(url, abr)]  abr: average bitrate
        video_candidates = []  # [(url, width)]
        with YoutubeDL(ytdl_opts) as inner:
            try:
                info = inner.extract_info(url, download=False)
            except DownloadError as e:  # noqa
                logger.warning(f"extract_info failed for {url}")
                info = None
            if info:
                for f in info['formats']:
                    if f.get('acodec', 'none') not in ('none', None):
                        audio_candidates.append((f['url'], f['abr']))
                    if (
                        f.get('vcodec', 'none') not in ('none', None)
                        and f.get('protocol', '') in ('https', 'http')
                    ):
                        video_candidates.append((f['url'], f['width']))
            if not (audio_candidates and video_candidates):
                return None
            audio_candidates = sorted(
                audio_candidates, key=lambda c: c[1] or 0, reverse=True
            )
            video_candidates = sorted(
                video_candidates, key=lambda c: c[1] or 0, reverse=True
            )
            # always use the best audio(with highest bitrate)
            audio_url = audio_candidates[0][0]
            # TODO: use policy on video because high-quality video may be slow
            video_url = video_candidates[0][0]
            return Media(VideoAudioManifest(video_url, audio_url), http_proxy=http_proxy)


if __name__ == '__main__':
    # A E2E test case for YTDL, this should print a playable url.
    rules = [
        {
            'name': 'match_source',
            'pattern': 'ytmusic',
            'http_proxy': 'http://127.0.0.1:7890'
        },
    ]

    ytdl = Ytdl(rules)
    url = 'https://music.youtube.com/watch?v=vKwowKeEv5w'

    # media = ytdl.select_audio(url, None, 'ytmusic')
    # print(media.url)
    # print()

    media = ytdl.select_video(url, None, 'ytmusic')
    assert media is not None
    assert isinstance(media.manifest, VideoAudioManifest)
    print(media.manifest.video_url)
    print(media.manifest.audio_url)



# mypy: disable-error-code=type-abstract
import logging
import warnings
from collections import Counter
from typing import Optional, TypeVar, List, TYPE_CHECKING

from feeluown.media import Media, MediaType
from feeluown.utils.aio import run_fn, as_completed
from feeluown.utils.dispatch import Signal
from feeluown.library.base import SearchType, ModelType
from feeluown.library.provider import Provider
from feeluown.library.excs import (
    MediaNotFound,
    ProviderAlreadyExists,
    ModelNotFound,
    ResourceNotFound,
)
from feeluown.library.flags import Flags as PF
from feeluown.library.models import (
    ModelFlags as MF,
    BaseModel,
    SimpleSearchResult,
    BriefVideoModel,
    BriefSongModel,
    SongModel,
    LyricModel,
    VideoModel,
    BriefAlbumModel,
    BriefArtistModel,
    AlbumModel,
)
from feeluown.library.model_state import ModelState
from feeluown.library.provider_protocol import (
    check_flag as check_flag_impl,
    SupportsSongLyric,
    SupportsSongMV,
    SupportsSongMultiQuality,
    SupportsVideoMultiQuality,
    SupportsSongWebUrl,
    SupportsVideoWebUrl,
    SupportsAlbumSongsReader,
    SupportsUserAutoLogin,
    SupportsImgUrlToMedia,
)
from feeluown.library.standby import (
    get_standby_score,
    STANDBY_DEFAULT_MIN_SCORE,
    STANDBY_FULL_SCORE,
)

if TYPE_CHECKING:
    from feeluown.ai import AI
    from feeluown.library.ytdl import Ytdl

logger = logging.getLogger(__name__)

T_p = TypeVar("T_p")


def raise_(e):
    raise e


class Library:
    """Resource entrypoints."""

    def __init__(self, providers_standby=None, enable_ai_standby_matcher=True):
        """

        :type app: feeluown.app.App
        """
        self._providers_standby = providers_standby
        self._providers = set()
        self.ytdl: Optional["Ytdl"] = None
        self.ai: Optional["AI"] = None

        self.provider_added = Signal()  # emit(AbstractProvider)
        self.provider_removed = Signal()  # emit(AbstractProvider)
        # TODO: implement this feature
        self.enable_ai_standby_matcher = enable_ai_standby_matcher

    def setup_ytdl(self, *args, **kwargs):
        from .ytdl import Ytdl

        self.ytdl = Ytdl(*args, **kwargs)

    def setup_ai(self, ai):
        self.ai = ai

    def register(self, provider):
        """register provider

        :raises ProviderAlreadyExists:
        :raises ValueError:
        """
        if not isinstance(provider, Provider):
            raise ValueError("invalid provider instance")
        for _provider in self._providers:
            if _provider.identifier == provider.identifier:
                raise ProviderAlreadyExists
        self._providers.add(provider)
        self.provider_added.emit(provider)

        if isinstance(provider, SupportsUserAutoLogin):
            logger.debug(f"Auto login for {provider.identifier} ...")
            run_fn(provider.auto_login)

    def deregister(self, provider) -> bool:
        """deregister provider

        .. versionchanged:: 4.0
           Do not raise exception anymore, return False instead.
        """
        if provider in self._providers:
            self._providers.remove(provider)
            self.provider_removed.emit(provider)
            return True
        return False

    def get(self, identifier) -> Optional[Provider]:
        """Obtain the provider instance by the resource provider’s unique identifier."""
        for provider in self._providers:
            if provider.identifier == identifier:
                return provider
        return None

    def list(self) -> List[Provider]:
        """List all resource providers."""
        return list(self._providers)

    def _filter(self, identifier_in=None):
        if identifier_in is None:
            return iter(self._providers)
        return filter(lambda p: p.identifier in identifier_in, self.list())

    def search(self, keyword, type_in=None, source_in=None, **kwargs):
        """search song/artist/album/playlist by keyword

        please use a_search method if you can.

        :param keyword: search keyword
        :param type_in: search type
        :param source_in: None or provider identifier list

        - TODO: support search with filters(by artist or by source)
        """
        type_in = SearchType.batch_parse(type_in) if type_in else [SearchType.so]
        for provider in self._filter(identifier_in=source_in):
            for type_ in type_in:
                try:
                    result = provider.search(keyword=keyword, type_=type_, **kwargs)
                except Exception:  # pylint: disable=broad-except
                    logger.exception("Search %s in %s failed.", keyword, provider)
                else:
                    if result is not None:
                        yield result

    async def a_search(
        self, keyword, source_in=None, timeout=None, type_in=None, return_err=False, **_
    ):
        """async version of search

        .. versionchanged:: 4.1.9
            Add `return_err` parameter.

        TODO: add Happy Eyeballs requesting strategy if needed
        """
        type_in = SearchType.batch_parse(type_in) if type_in else [SearchType.so]

        # Wrap the search function to associate the result with source.
        def wrap_search(pvd, kw, t):
            def search():
                try:
                    res = pvd.search(kw, type_=t)
                except Exception as e:  # noqa
                    if return_err:
                        logger.exception("One provider search failed")
                        return SimpleSearchResult(
                            q=keyword,
                            source=pvd.identifier,  # noqa
                            err_msg=f"{type(e)}",
                        )
                    raise e
                # When a provider does not implement search method, it returns None.
                if res is not None and (
                    res.songs
                    or res.albums
                    or res.artists
                    or res.videos
                    or res.playlists
                ):
                    return res
                # Pass empty err_msg, otherwise `track-search-result-empty`
                # can be ignored.
                return SimpleSearchResult(q=keyword, source=pvd.identifier, err_msg="")

            return search

        fs = []  # future list
        for provider in self._filter(identifier_in=source_in):
            for type_ in type_in:
                future = run_fn(wrap_search(provider, keyword, type_))
                fs.append(future)
        for task_ in as_completed(fs, timeout=timeout):
            try:
                result = await task_
            except Exception as e:  # noqa
                logger.exception("One search task failed due to asyncio")
            else:
                yield result

    async def a_song_prepare_media_no_exc(self, standby, policy):
        media = None
        try:
            media = await run_fn(self.song_prepare_media, standby, policy)
        except MediaNotFound as e:
            logger.debug(f"standby media not found: {e}")
        except:  # noqa
            logger.exception(f"get standby:{standby} media failed")
        return media

    async def a_list_song_standby_v2(
        self,
        song,
        audio_select_policy=">>>",
        source_in=None,
        score_fn=None,
        min_score=STANDBY_DEFAULT_MIN_SCORE,
        limit=1,
    ):
        """list song standbys and their media

        .. versionadded:: 3.7.8
        """
        if source_in is None:
            pvd_ids = self._providers_standby or [pvd.identifier for pvd in self.list()]
        else:
            pvd_ids = [pvd.identifier for pvd in self._filter(identifier_in=source_in)]
        if score_fn is None:
            score_fn = get_standby_score
        limit = max(limit, 1)

        q = "{} {}".format(song.title_display, song.artists_name_display)
        standby_score_list = []  # [(standby, score), (standby, score)]
        song_media_list = []  # [(standby, media), (standby, media)]
        top2_standby = []
        async for result in self.a_search(q, source_in=pvd_ids):
            if result is None:
                continue
            # Only check the first 3 songs
            for i, standby in enumerate(result.songs):
                # HACK(cosven): I think the local provider should not be included,
                #   because the search algorithm of local provider is so bad.
                if i < 2 and standby.source != "local":
                    top2_standby.append(standby)
                score = score_fn(song, standby)
                if score == STANDBY_FULL_SCORE:
                    media = await self.a_song_prepare_media_no_exc(
                        standby, audio_select_policy
                    )
                    if media is None:
                        continue
                    logger.info(f"Find full score standby for song:{q}")
                    song_media_list.append((standby, media))
                    if len(song_media_list) >= limit:
                        # Return as early as possible to get better performance
                        return song_media_list
                elif score >= min_score:
                    standby_score_list.append((standby, score))
        if standby_score_list:
            standby_pvd_id_set = {standby.source for standby, _ in standby_score_list}
            logger.info(
                f"Find {len(standby_score_list)} similar songs "
                f"from {','.join(standby_pvd_id_set)}. Try to get a valid media"
            )
            max_per_source = 2
            standby_score_list_2 = []
            counter = Counter()
            for s, score in standby_score_list:
                if counter[s.source] >= max_per_source:
                    continue
                counter[s.source] += 1
                standby_score_list_2.append((s, score))

            assert len(standby_score_list_2) <= max_per_source * len(standby_pvd_id_set)
            sorted_standby_score_list = sorted(
                standby_score_list_2,
                key=lambda song_score: song_score[1],
                reverse=True,
            )
            for standby, _ in sorted_standby_score_list:
                # TODO: send multiple requests at a time.
                media = await self.a_song_prepare_media_no_exc(
                    standby, audio_select_policy
                )
                if media is not None:
                    song_media_list.append((standby, media))
                    if len(song_media_list) >= limit:
                        return song_media_list
            if song_media_list:
                return song_media_list
        return song_media_list

    def check_flags(self, source: str, model_type: ModelType, flags: PF) -> bool:
        """Check if a provider satisfies the specific ability for a model type

        .. note::

             Currently, we use ProviderFlags to define which ability a
             provider has. In the future, we may use typing.Protocol.
             So you should use :meth:`check_flags` method to check ability
             instead of compare provider flags directly.
        """
        provider = self.get(source)
        if provider is None:
            return False

        # Check each flag.
        for value in PF.__members__.values():
            if value in PF(flags):
                if check_flag_impl(provider, model_type, flags) is False:
                    return False
        return True

    def check_flags_by_model(self, model: BaseModel, flags: PF) -> bool:
        """Alias for check_flags"""
        warnings.warn("please use isinstance(provider, protocol_cls)")
        return self.check_flags(model.source, ModelType(model.meta.model_type), flags)

    # -----
    # Songs
    # -----
    def song_upgrade(self, song: BriefSongModel) -> SongModel:
        return self._model_upgrade(song)  # type: ignore

    def song_prepare_media(self, song: BriefSongModel, policy) -> Media:
        provider = self.get(song.source)
        if provider is None:
            raise MediaNotFound(f"provider({song.source}) not found")
        media = None
        if isinstance(provider, SupportsSongMultiQuality):
            try:
                media, _ = provider.song_select_media(song, policy)
            except MediaNotFound as e:
                if e.reason is MediaNotFound.Reason.check_children:
                    raise
                else:
                    media = None
        if not media:
            if self.ytdl is not None and isinstance(provider, SupportsSongWebUrl):
                song_web_url = provider.song_get_web_url(song)
                media = self.ytdl.select_audio(song_web_url, policy, source=song.source)
                found = media is not None
                logger.debug(
                    f"ytdl select audio for {song_web_url} finished, found: {found}"
                )
            if not media:
                raise MediaNotFound("provider returns empty media")
        return media

    def song_get_web_url(self, song: BriefSongModel) -> str:
        """Get song web url

        :raises ResourceNotFound: provider/song/web_url is not found
        :raises ProviderIOError: provider raises error during get_web_url

        .. versionadded:: 4.1.8
        """
        provider = self.get(song.source)
        if provider is None:
            raise ResourceNotFound(f"provider({song.source}) not found")
        if isinstance(provider, SupportsSongWebUrl):
            return provider.song_get_web_url(song)
        raise ResourceNotFound(reason=ResourceNotFound.Reason.not_supported)

    def song_prepare_mv_media(self, song: BriefSongModel, policy) -> Media:
        """

        .. versionadded:: 3.7.5
        """
        mv = self.song_get_mv(song)
        if mv is not None:
            media = self.video_prepare_media(mv, policy)
            return media
        raise MediaNotFound("provider returns empty media")

    def song_get_mv(self, song: BriefSongModel) -> Optional[VideoModel]:
        """Get the MV model of a song."""
        provider = self.get(song.source)
        if isinstance(provider, SupportsSongMV):
            return provider.song_get_mv(song)
        return None

    def song_get_lyric(self, song: BriefSongModel) -> Optional[LyricModel]:
        """Get the lyric model of a song.

        Return None when lyric does not exist instead of raising exceptions,
        because it is predictable.
        """
        provider = self.get(song.source)
        if isinstance(provider, SupportsSongLyric):
            return provider.song_get_lyric(song)
        return None

    # --------
    # Album
    # --------
    def album_upgrade(self, album: BriefAlbumModel):
        return self._model_upgrade(album)

    def album_list_songs(self, album: BriefAlbumModel):
        """
        :raises ResourceNotFound:
        :raises ProviderIOError:

        .. versionadded:: 4.1.11
        """
        if not isinstance(album, AlbumModel):
            ualbum = self.album_upgrade(album)
        else:
            ualbum = album
        if ualbum.song_count == 0:
            return []
        if ualbum.songs:
            return ualbum.songs
        provider = self.get(ualbum.source)
        if isinstance(provider, SupportsAlbumSongsReader):
            return list(provider.album_create_songs_rd(ualbum))
        raise ResourceNotFound(reason=ResourceNotFound.Reason.not_supported)

    # --------
    # Artist
    # --------
    def artist_upgrade(self, artist: BriefArtistModel):
        return self._model_upgrade(artist)

    # --------
    # Playlist
    # --------

    def playlist_upgrade(self, playlist):
        return self._model_upgrade(playlist)

    # -------------------------
    # generic methods for model
    # -------------------------
    def model_get(self, pid, mtype, mid):
        """Get a (normal) model instance.

        :param pid: provider id
        :param mtype: model type
        :param mid: model id
        :return: model

        :raise ResourceNotFound: model does not exist
        """
        provider = self.get(pid)
        if provider is None:
            raise ModelNotFound(f"provider:{pid} not found")
        return provider.model_get(mtype, mid)

    def model_get_cover(self, model):
        """Get the cover url of model

        :param model: model which has a 'cover' field.
        :return: cover url if exists, else ''.
        """
        if MF.normal not in model.meta.flags:
            try:
                um = self._model_upgrade(model)
            except ResourceNotFound:
                return ""
        else:
            um = model
        if ModelType(model.meta.model_type) is ModelType.artist:
            cover = um.pic_url
        elif ModelType(model.meta.model_type) is ModelType.song:
            cover = um.pic_url
            if not cover:
                album = um.album
                return self.model_get_cover(album)
        else:
            cover = um.cover
        return cover

    def model_get_cover_media(self, model) -> Optional[Media]:
        """Get cover media and preserve provider-level network options."""
        cover_url = self.model_get_cover(model)
        if not cover_url:
            return None
        provider = self.get(model.source)
        if isinstance(provider, SupportsImgUrlToMedia):
            return provider.img_url_to_media(cover_url)
        return Media(cover_url, MediaType.image)

    def _model_upgrade(self, model):
        """Upgrade a model to normal model.

        :raises ModelNotFound: the model does not exist

        Note you may catch ResourceNotFound exception to simplify your code.

        .. versionchanged:: 3.8.11
            Raise ModelNotFound if the model does not exist.
            Before ModelCannotUpgrade was raised.
        """
        # Return model directly if it is already a normal(upgraded) model.
        if MF.normal in model.meta.flags:
            return model

        model_type = ModelType(model.meta.model_type)
        provider = self.get(model.source)
        if provider is None:
            raise ModelNotFound(f"provider:{model.source} not found")
        try:
            upgraded_model = provider.model_get(model_type, model.identifier)
        except ModelNotFound as e:
            if e.reason is ModelNotFound.Reason.not_found:
                model.state = ModelState.not_exists
            elif e.reason is ModelNotFound.Reason.not_supported:
                model.state = ModelState.cant_upgrade
            raise
        if upgraded_model is None:  # some provider does not implement
            raise ModelNotFound(f"{provider} implementation error, it returns None :(")
        return upgraded_model

    # --------
    # Video
    # --------
    def video_upgrade(self, video):
        return self._model_upgrade(video)

    def video_prepare_media(self, video: BriefVideoModel, policy) -> Media:
        """Prepare media for video.

        :param video: either a v1 MvModel or a v2 (Brief)VideoModel.
        """
        provider = self.get(video.source)
        try:
            if isinstance(provider, SupportsVideoMultiQuality):
                media, _ = provider.video_select_media(video, policy)
                if not media:
                    raise MediaNotFound("provider returns empty media")
            else:
                raise MediaNotFound("provider or video not found")
        except MediaNotFound:
            if self.ytdl is not None and isinstance(provider, SupportsVideoWebUrl):
                video_web_url = provider.video_get_web_url(video)
                media = self.ytdl.select_video(
                    video_web_url, policy, source=video.source
                )
                found = media is not None
                logger.debug(
                    f"ytdl select video for {video_web_url} finished, found: {found}"
                )
            else:
                media = None
            if not media:
                raise
        return media
